from functools import wraps
from typing import Callable, Optional, TypeVar


C = TypeVar("C", bound=Callable)


def bindable(bindable: C) -> C:
    """
    Mark a method as a "bindable" method.

    If a L{Klein.app} resource is found on an instance object (i.e. is returned
    from C{YourObject().app.resource()}), it will pass C{self} from that
    instance to all of its routes, making a signature of 2 arguments: C{self}
    and C{request} However, if it's found globally (i.e. C{app = Klein()};
    C{@app.route(...)} at global scope), then it will only receive one:
    C{request}.  However, for decorators that must be able to live between
    C{@route} and the user's function, but still need to manipulate the
    C{request} object, they need to be invoked with a consistent argument
    signature.  A method decorated with C{@bindable} will therefore always take
    C{instance, request} as its first two arguments, even if C{instance} is
    C{None} when the L{Klein} object is not bound to an instance.

    @return: its argument, modified to mark it as unconditionally requiring an
        instance argument.
    """
    bindable.__klein_bound__ = True  # type: ignore[attr-defined]
    return bindable


def modified(
    modification: str,
    original: Callable,
    modifier: Optional[Callable] = None,
) -> Callable:
    """
    Annotate a callable as a modified wrapper of an original callable.

    @param modification: A name for the type of modification, for example "form
        processor" or "request forwarder"; this will be tacked on to the name
        of the resulting function.

    @param modifier: Another decorator which, if given, will be applied to the
        function that decorates this function.  Additionally, I{any new
        attributes} set on the decorated function by C{modifier} will be
        I{copied to} C{original}.  This allows attributes set by "inner"
        decorators such as L{klein.Form.handler} and L{klein.app.Klein.route}
        to set attributes that will be visible at the top level.

    @return: A new callable; this may have a different argument signature or
        return value, and is only related to C{original} in the sense that it
        likely calls it.
    """

    def decorator(wrapper: Callable[..., C]) -> Callable[..., C]:
        result = named(modification + " for " + original.__name__)(
            wraps(original)(wrapper)
        )
        result.__original__ = original  # type: ignore[attr-defined]
        if modifier is not None:
            before = set(wrapper.__dict__.keys())
            result = modifier(result)
            after = set(wrapper.__dict__.keys())
            for key in after - before:
                setattr(original, key, wrapper.__dict__[key])
        return result

    return decorator


def named(name: str) -> Callable[[C], C]:
    """
    Change the name of a function to the given name.
    """

    def decorator(original: C) -> C:
        original.__name__ = str(name)
        original.__qualname__ = str(name)
        return original

    return decorator


def originalName(function: Callable) -> str:
    """
    Get the original, user-specified name of C{function}, chasing back any
    wrappers applied with C{modified}.
    """
    fnext: Optional[Callable] = function
    while fnext is not None:
        function = fnext
        fnext = getattr(function, "__original__", None)
    return function.__name__



# -*- test-case-name: klein.test.test_headers -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Support for interoperability with L{twisted.web.http_headers.Headers}.
"""

from typing import AnyStr, Iterable, Tuple, cast

from attr import attrib, attrs
from attr.validators import instance_of
from zope.interface import implementer

from twisted.web.http_headers import Headers

from ._headers import (
    IMutableHTTPHeaders,
    RawHeaders,
    String,
    headerNameAsBytes,
    headerValueAsText,
    normalizeHeaderName,
    rawHeaderName,
    rawHeaderNameAndValue,
)


__all__ = ()


@implementer(IMutableHTTPHeaders)
@attrs(frozen=True)
class HTTPHeadersWrappingHeaders:
    """
    HTTP entity headers.

    This is an L{IMutableHTTPHeaders} implementation that wraps a L{Headers}
    object.
    """

    # NOTE: In case Headers has different ideas about encoding text than we do,
    # always interact with it using bytes, not str.

    _headers: Headers = attrib(validator=instance_of(Headers))

    @property
    def rawHeaders(self) -> RawHeaders:
        def pairs() -> Iterable[Tuple[bytes, bytes]]:
            for name, values in self._headers.getAllRawHeaders():
                name = normalizeHeaderName(name)
                for value in values:
                    yield (name, value)

        return tuple(pairs())

    def getValues(self, name: AnyStr) -> Iterable[AnyStr]:
        if isinstance(name, bytes):
            values = cast(
                Iterable[AnyStr], self._headers.getRawHeaders(name, default=())
            )
        elif isinstance(name, str):
            # typing note: getRawHeaders is typed to return an optional even if
            # default is not None. We could assert not None, but are casting
            # here so that if the hints for getRawHeaders are fixed later,
            # mypy will tell us to remove the useless cast.
            values = cast(
                Iterable[str],
                self._headers.getRawHeaders(
                    headerNameAsBytes(name), default=()
                ),
            )
            values = (headerValueAsText(value) for value in values)
        else:
            raise TypeError(f"name {name!r} must be str or bytes")

        return values

    def remove(self, name: String) -> None:
        self._headers.removeHeader(rawHeaderName(name))

    def addValue(self, name: AnyStr, value: AnyStr) -> None:
        rawName, rawValue = rawHeaderNameAndValue(name, value)

        self._headers.addRawHeader(rawName, rawValue)



"""
Dependency-Injected HTTP metadata.
"""

from typing import Any, Dict, Mapping, Sequence, Type, Union, cast

import attr
from hyperlink import DecodedURL
from zope.interface import Interface, implementer, provider

from twisted.python.components import Componentized
from twisted.web.iweb import IRequest

from .interfaces import (
    IDependencyInjector,
    IRequestLifecycle,
    IRequiredParameter,
)


def urlFromRequest(request: IRequest) -> DecodedURL:
    sentHeader = request.getHeader(b"host")
    if sentHeader is not None:
        sentHeader = sentHeader.decode("charmap")
        if ":" in sentHeader:
            host, port = sentHeader.split(":")
            port = int(port)
        else:
            host = sentHeader
            port = None
    else:
        client = request.client  # type: ignore[attr-defined]
        host = client.host
        port = client.port

    url = DecodedURL.fromText(request.uri.decode("charmap"))
    url = url.replace(
        scheme="https" if request.isSecure() else "http",
        host=host,
        port=port,
    )
    return url


@provider(IRequiredParameter, IDependencyInjector)
class RequestURL:
    """
    Require a hyperlink L{DecodedURL} object from a L{Requirer}.

    @since: Klein 23.12.0
    """

    @classmethod
    def registerInjector(
        cls,
        injectionComponents: Componentized,
        parameterName: str,
        requestLifecycle: IRequestLifecycle,
    ) -> IDependencyInjector:
        # typing note: https://github.com/Shoobx/mypy-zope/issues/39
        return cast(IDependencyInjector, cls())

    @classmethod
    def injectValue(
        cls,
        instance: Any,
        request: IRequest,
        routeParams: Dict[str, Any],
    ) -> DecodedURL:
        return urlFromRequest(request)

    @classmethod
    def finalize(cls) -> None:
        "Nothing to do upon finalization."


@implementer(IRequiredParameter, IDependencyInjector)
@attr.frozen
class RequestComponent:
    """
    Require a hyperlink L{DecodedURL} object from a L{Requirer}.

    @since: Klein 23.12.0
    """

    interface: Type[Interface]

    def registerInjector(
        self,
        injectionComponents: Componentized,
        parameterName: str,
        requestLifecycle: IRequestLifecycle,
    ) -> IDependencyInjector:
        return self

    def injectValue(
        self, instance: Any, request: IRequest, routeParams: Dict[str, Any]
    ) -> DecodedURL:
        return cast(
            DecodedURL,
            cast(Componentized, request).getComponent(self.interface),
        )

    def finalize(cls) -> None:
        "Nothing to do upon finalization."


@attr.s(auto_attribs=True, frozen=True)
class Response:
    """
    Metadata about an HTTP response, with an object that Klein knows how to
    understand.

    This includes:

        - an HTTP response code

        - some HTTP headers

        - a body object, which can be anything else Klein understands; for
          example, an IResource, an IRenderable, str, bytes, etc.

    @since: Klein 23.12.0
    """

    code: int = 200
    headers: Mapping[
        Union[str, bytes], Union[str, bytes, Sequence[Union[str, bytes]]]
    ] = attr.ib(factory=dict)
    body: Any = ""

    def _applyToRequest(self, request: IRequest) -> Any:
        """
        Apply this L{Response} to the given L{IRequest}, setting its response
        code and headers.

        Private because:

            - this should only ever be applied by Klein, and

            - hopefully someday soon this will be replaced with something that
              actually creates a txrequest-style response object.
        """
        request.setResponseCode(self.code)
        for headerName, headerValueOrValues in self.headers.items():
            if not isinstance(headerValueOrValues, (str, bytes)):
                headerValues = headerValueOrValues
            else:
                headerValues = [headerValueOrValues]
            request.responseHeaders.setRawHeaders(headerName, headerValues)
        return self.body



from typing import Any, Callable, Dict, Generator, List, Sequence, Type

import attr
from zope.interface import Interface, implementer

from twisted.internet.defer import inlineCallbacks
from twisted.python.components import Componentized
from twisted.web.iweb import IRequest

from ._app import _call
from ._decorators import bindable, modified
from .interfaces import (
    EarlyExit,
    IDependencyInjector,
    IRequestLifecycle,
    IRequiredParameter,
)


@implementer(IRequestLifecycle)
@attr.s(auto_attribs=True)
class RequestLifecycle:
    """
    Mechanism to run hooks at the start of a request managed by a L{Requirer}.
    """

    _prepareHooks: List = attr.ib(factory=list)

    def addPrepareHook(
        self,
        beforeHook: Callable,
        requires: Sequence[Type[Interface]] = (),
        provides: Sequence[Type[Interface]] = (),
    ) -> None:
        # TODO: topological requirements sort
        self._prepareHooks.append(beforeHook)

    @inlineCallbacks
    def runPrepareHooks(
        self, instance: Any, request: IRequest
    ) -> Generator[Any, object, None]:
        """
        Execute all the hooks added with L{RequestLifecycle.addPrepareHook}.
        This is invoked by the L{requires} route machinery.

        @param instance: The instance bound to the Klein route.

        @param request: The IRequest being processed.
        """
        for hook in self._prepareHooks:
            yield _call(instance, hook, request)


_routeDecorator = Any  # a decorator like @route
_routeT = Any  # a thing decorated by a decorator like @route

_prerequisiteCallback = Callable[[IRequestLifecycle], None]


@attr.s(auto_attribs=True)
class Requirer:
    """
    Dependency injection for required parameters.
    """

    _prerequisites: List[_prerequisiteCallback] = attr.ib(factory=list)

    def prerequisite(
        self,
        providesComponents: Sequence[Type[Interface]],
        requiresComponents: Sequence[Type[Interface]] = (),
    ) -> Callable[[Callable], Callable]:
        """
        Specify a component that is a pre-requisite of every request routed
        through this requirer's C{require} method.  Used like so::

            requirer = Requirer()

            @requirer.prerequisite([IFoo])
            @inlineCallbacks
            def fooForRequest(request):
                request.setComponent(IFoo, someFooComponent)

        @note: C{requiresComponents} is, at this point, for the reader's
            interest only, the framework will not topologically sort
            dependencies; you must presently register prerequisites in the
            order you want them to be called.
        """

        def decorator(prerequisiteMethod: Callable) -> Callable:
            def oneHook(lifecycle: IRequestLifecycle) -> None:
                lifecycle.addPrepareHook(
                    prerequisiteMethod,
                    requires=requiresComponents,
                    provides=providesComponents,
                )

            self._prerequisites.append(oneHook)
            return prerequisiteMethod

        return decorator

    def require(
        self, routeDecorator: _routeT, **requiredParameters: IRequiredParameter
    ) -> _routeDecorator:
        """
        Inject the given dependencies while running the given route.
        """

        def decorator(functionWithRequirements: Callable) -> Callable:
            injectionComponents = Componentized()
            lifecycle = RequestLifecycle()
            injectionComponents.setComponent(IRequestLifecycle, lifecycle)

            injectors: Dict[str, IDependencyInjector] = {}

            for parameterName, required in requiredParameters.items():
                injectors[parameterName] = required.registerInjector(
                    injectionComponents, parameterName, lifecycle
                )

            for prereq in self._prerequisites:
                prereq(lifecycle)

            for v in injectors.values():
                v.finalize()

            @modified("dependency-injecting route", functionWithRequirements)
            @bindable
            @inlineCallbacks
            def router(
                instance: Any, request: IRequest, *args: Any, **routeParams: Any
            ) -> Any:
                injected = routeParams.copy()
                try:
                    yield lifecycle.runPrepareHooks(instance, request)
                    for k, injector in injectors.items():
                        injected[k] = yield injector.injectValue(
                            instance, request, routeParams
                        )
                except EarlyExit as ee:
                    result = ee.alternateReturnValue
                else:
                    result = yield _call(
                        instance, functionWithRequirements, *args, **injected
                    )
                return result

            fWR, iC = functionWithRequirements, injectionComponents
            fWR.injectionComponents = iC  # type: ignore[attr-defined]
            routeDecorator(router)
            return functionWithRequirements

        return decorator



# -*- test-case-name: klein.test.test_headers_compat -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._headers}.
"""

from twisted.web.http_headers import Headers

from .._headers import (
    IMutableHTTPHeaders,
    RawHeaders,
    normalizeRawHeadersFrozen,
)
from .._headers_compat import HTTPHeadersWrappingHeaders
from ._trial import TestCase
from .test_headers import MutableHTTPHeadersTestsMixIn


try:
    from twisted.web.http_headers import _sanitizeLinearWhitespace
except ImportError:  # pragma: no cover
    _sanitizeLinearWhitespace = None  # type: ignore[assignment]


def _twistedHeaderNormalize(value: str) -> str:
    """
    Normalize the given header value according to the rules of the installed
    Twisted version.
    """
    if _sanitizeLinearWhitespace is None:  # pragma: no cover
        return value  # type: ignore[unreachable]
    else:
        return _sanitizeLinearWhitespace(value.encode("utf-8")).decode("utf-8")


__all__ = ()


class HTTPHeadersWrappingHeadersTests(MutableHTTPHeadersTestsMixIn, TestCase):
    """
    Tests for L{HTTPHeadersWrappingHeaders}.
    """

    def assertRawHeadersEqual(
        self, rawHeaders1: RawHeaders, rawHeaders2: RawHeaders
    ) -> None:
        super().assertRawHeadersEqual(sorted(rawHeaders1), sorted(rawHeaders2))

    def headerNormalize(self, value: str) -> str:
        return _twistedHeaderNormalize(value)

    def headers(self, rawHeaders: RawHeaders) -> IMutableHTTPHeaders:
        headers = Headers()
        for rawName, rawValue in rawHeaders:
            headers.addRawHeader(rawName, rawValue)

        return HTTPHeadersWrappingHeaders(headers=headers)

    def test_rawHeaders(self) -> None:
        """
        L{MutableHTTPHeaders.rawHeaders} equals raw headers matching the
        L{Headers} given at init time.
        """
        rawHeaders = ((b"b", b"2a"), (b"a", b"1"), (b"B", b"2b"))
        webHeaders = Headers()
        for name, value in rawHeaders:
            webHeaders.addRawHeader(name, value)
        headers = HTTPHeadersWrappingHeaders(headers=webHeaders)

        # Note that Headers does not give you back header names in network
        # order, but it should give us back values in network order.
        # So we need to normalize our way around.

        normalizedRawHeaders = normalizeRawHeadersFrozen(rawHeaders)

        self.assertEqual(
            sorted(headers.rawHeaders), sorted(normalizedRawHeaders)
        )



"""
Shared tools for Klein's test suite.
"""

from abc import ABC, abstractmethod
from typing import Type, TypeVar, cast

from twisted.internet.defer import Deferred
from twisted.trial.unittest import SynchronousTestCase


class EqualityTestsMixin(ABC):
    """
    A mixin defining tests for the standard implementation of C{==} and C{!=}.
    """

    @abstractmethod
    def anInstance(self) -> object:
        """
        Return an instance of the class under test.  Each call to this method
        must return a different object.  All objects returned must be equal to
        each other.
        """

    @abstractmethod
    def anotherInstance(self) -> object:
        """
        Return an instance of the class under test.  Each call to this method
        must return a different object.  The objects must not be equal to the
        objects returned by C{anInstance}.  They may or may not be equal to
        each other (they will not be compared against each other).
        """

    def test_identicalEq(self) -> None:
        """
        An object compares equal to itself using the C{==} operator.
        """
        o = self.anInstance()
        cast(SynchronousTestCase, self).assertTrue(o == o)

    def test_identicalNe(self) -> None:
        """
        An object doesn't compare not equal to itself using the C{!=} operator.
        """
        o = self.anInstance()
        cast(SynchronousTestCase, self).assertFalse(o != o)

    def test_sameEq(self) -> None:
        """
        Two objects that are equal to each other compare equal to each other
        using the C{==} operator.
        """
        a = self.anInstance()
        b = self.anInstance()
        cast(SynchronousTestCase, self).assertTrue(a == b)

    def test_sameNe(self) -> None:
        """
        Two objects that are equal to each other do not compare not equal to
        each other using the C{!=} operator.
        """
        a = self.anInstance()
        b = self.anInstance()
        cast(SynchronousTestCase, self).assertFalse(a != b)

    def test_differentEq(self) -> None:
        """
        Two objects that are not equal to each other do not compare equal to
        each other using the C{==} operator.
        """
        a = self.anInstance()
        b = self.anotherInstance()
        cast(SynchronousTestCase, self).assertFalse(a == b)

    def test_differentNe(self) -> None:
        """
        Two objects that are not equal to each other compare not equal to each
        other using the C{!=} operator.
        """
        a = self.anInstance()
        b = self.anotherInstance()
        cast(SynchronousTestCase, self).assertTrue(a != b)

    def test_anotherTypeEq(self) -> None:
        """
        The object does not compare equal to an object of an unrelated type
        (which does not implement the comparison) using the C{==} operator.
        """
        a = self.anInstance()
        b = object()
        cast(SynchronousTestCase, self).assertFalse(a == b)

    def test_anotherTypeNe(self) -> None:
        """
        The object compares not equal to an object of an unrelated type (which
        does not implement the comparison) using the C{!=} operator.
        """
        a = self.anInstance()
        b = object()
        cast(SynchronousTestCase, self).assertTrue(a != b)

    def test_delegatedEq(self) -> None:
        """
        The result of comparison using C{==} is delegated to the right-hand
        operand if it is of an unrelated type.
        """

        class Delegate:
            def __eq__(self, other: object) -> bool:
                # Do something crazy and easily identifiable.
                return cast(bool, [self])

        a = self.anInstance()
        b = Delegate()
        cast(SynchronousTestCase, self).assertEqual(a == b, [b])

    def test_delegateNe(self) -> None:
        """
        The result of comparison using C{!=} is delegated to the right-hand
        operand if it is of an unrelated type.
        """

        class Delegate:
            def __ne__(self, other: object) -> bool:
                # Do something crazy and easily identifiable.
                return cast(bool, [self])

        a = self.anInstance()
        b = Delegate()
        cast(SynchronousTestCase, self).assertEqual(a != b, [b])


_T = TypeVar("_T")


def recover(d: "Deferred[_T]", exc_type: Type[Exception]) -> "Deferred[_T]":
    return d.addErrback(lambda f: f.trap(exc_type))



# -*- test-case-name: klein.test.test_request -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
HTTP headers API.
"""

from typing import AnyStr, Iterable, Tuple, Union

from attr import Factory, attrib, attrs
from zope.interface import implementer

from ._imessage import (
    IHTTPHeaders,
    IMutableHTTPHeaders,
    MutableRawHeaders,
    RawHeader,
    RawHeaders,
)


__all__ = ()


String = Union[bytes, str]


# Encoding/decoding header data

HEADER_NAME_ENCODING = "iso-8859-1"
HEADER_VALUE_ENCODING = "iso-8859-1"


def headerNameAsBytes(name: String) -> bytes:
    """
    Convert a header name to bytes if necessary.
    """
    if isinstance(name, bytes):
        return name
    else:
        return name.encode(HEADER_NAME_ENCODING)


def headerNameAsText(name: String) -> str:
    """
    Convert a header name to str if necessary.
    """
    if isinstance(name, str):
        return name
    else:
        return name.decode(HEADER_NAME_ENCODING)


def headerValueAsBytes(value: String) -> bytes:
    """
    Convert a header value to bytes if necessary.
    """
    if isinstance(value, bytes):
        return value
    else:
        return value.encode(HEADER_VALUE_ENCODING)


def headerValueAsText(value: String) -> str:
    """
    Convert a header value to str if necessary.
    """
    if isinstance(value, str):
        return value
    else:
        return value.decode(HEADER_VALUE_ENCODING)


def normalizeHeaderName(name: AnyStr) -> AnyStr:
    """
    Normalize a header name.
    """
    return name.lower()


# Internal data representation


def normalizeRawHeaders(
    headerPairs: Iterable[Iterable[String]],
) -> Iterable[RawHeader]:
    for pair in headerPairs:
        try:
            name, value = pair
        except ValueError:
            raise ValueError("header pair must be a 2-item iterable")

        name = normalizeHeaderName(headerNameAsBytes(name))
        value = headerValueAsBytes(value)

        yield (normalizeHeaderName(name), value)


def normalizeRawHeadersFrozen(
    headerPairs: Iterable[Iterable[bytes]],
) -> RawHeaders:
    return tuple(normalizeRawHeaders(headerPairs))


def normalizeRawHeadersMutable(
    headerPairs: Iterable[Iterable[bytes]],
) -> MutableRawHeaders:
    return list(normalizeRawHeaders(headerPairs))


def getFromRawHeaders(rawHeaders: RawHeaders, name: AnyStr) -> Iterable[AnyStr]:
    """
    Get a value from raw headers.
    """
    if isinstance(name, bytes):
        name = normalizeHeaderName(name)
        return (v for n, v in rawHeaders if name == n)

    if isinstance(name, str):
        rawName = headerNameAsBytes(normalizeHeaderName(name))
        return (headerValueAsText(v) for n, v in rawHeaders if rawName == n)

    raise TypeError(f"name {name!r} must be str or bytes")


def rawHeaderName(name: String) -> bytes:
    if isinstance(name, bytes):
        return name
    elif isinstance(name, str):
        return headerNameAsBytes(name)
    else:
        raise TypeError(f"name {name!r} must be str or bytes")


def rawHeaderNameAndValue(name: String, value: String) -> Tuple[bytes, bytes]:
    if isinstance(name, bytes):
        if not isinstance(value, bytes):
            raise TypeError(
                "value {!r} must be bytes to match name {!r}".format(
                    value, name
                )
            )
        return (name, value)

    elif isinstance(name, str):
        if not isinstance(value, str):
            raise TypeError(
                f"value {value!r} must be str to match name {name!r}"
            )
        return (headerNameAsBytes(name), headerValueAsBytes(value))

    else:
        raise TypeError(f"name {name!r} must be str or bytes")


# Implementation


@implementer(IHTTPHeaders)
@attrs(frozen=True)
class FrozenHTTPHeaders:
    """
    Immutable HTTP entity headers.
    """

    rawHeaders: RawHeaders = attrib(
        converter=normalizeRawHeadersFrozen,
        default=(),
    )

    def getValues(self, name: AnyStr) -> Iterable[AnyStr]:
        return getFromRawHeaders(self.rawHeaders, name)


@implementer(IMutableHTTPHeaders)
@attrs(frozen=True)
class MutableHTTPHeaders:
    """
    Mutable HTTP entity headers.
    """

    _rawHeaders: MutableRawHeaders = attrib(
        converter=normalizeRawHeadersMutable,
        default=Factory(list),
    )

    @property
    def rawHeaders(self) -> RawHeaders:
        return tuple(self._rawHeaders)

    def getValues(self, name: AnyStr) -> Iterable[AnyStr]:
        return getFromRawHeaders(self._rawHeaders, name)

    def remove(self, name: String) -> None:
        rawName = rawHeaderName(name)

        self._rawHeaders[:] = [p for p in self._rawHeaders if p[0] != rawName]

    def addValue(self, name: AnyStr, value: AnyStr) -> None:
        self._rawHeaders.append(rawHeaderNameAndValue(name, value))



# -*- test-case-name: klein.test.test_resource.GlobalAppTests -*-

"""
This module, L{klein.resource}, serves two purposes:

    - It's the global C{resource()} method on the global L{klein.Klein}
      application.

    - It's the module where L{KleinResource} is defined.
"""

from sys import modules
from typing import Any, Callable, Union

from ._app import resource as _globalResourceMethod
from ._resource import KleinResource as _KleinResource
from ._resource import ensure_utf8_bytes


KleinResource = _KleinResource


class _SpecialModuleObject:
    """
    See the test in
    L{klein.test.test_resource.GlobalAppTests.test_weird_resource_situation}
    for an explanation.
    """

    __all__ = (
        "KleinResource",
        "ensure_utf8_bytes",
    )

    KleinResource = _KleinResource

    def __init__(self, preserve: Any) -> None:
        self.__preserve__ = preserve

    @property
    def ensure_utf8_bytes(self) -> Callable[[Union[str, bytes]], bytes]:
        return ensure_utf8_bytes

    def __call__(self) -> _KleinResource:
        """
        Return an L{IResource} which suitably wraps this app.

        @returns: An L{IResource}
        """
        # Use the same docstring as the real implementation to reduce
        # confusion.
        return _globalResourceMethod()

    def __repr__(self) -> str:
        """
        Give a special C{repr()} to make the dual purpose of this object clear.
        """
        return "<special bound method/module klein.resource>"


module = _SpecialModuleObject(modules[__name__])
modules[__name__] = module  # type: ignore[assignment]



"""
Since we support a range of Python and Mypy versions where certain features are
available across L{typing} and L{typing_extensions}, we put those aliases here
to avoid repeating conditional import logic.
"""
import sys


if sys.version_info > (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


if sys.version_info > (3, 10):
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec


__all__ = [
    "Protocol",
    "ParamSpec",
    "Concatenate",
]



# -*- test-case-name: klein.test.test_headers -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._headers}.
"""

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import (
    AnyStr,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from .._headers import (
    HEADER_NAME_ENCODING,
    HEADER_VALUE_ENCODING,
    FrozenHTTPHeaders,
    IHTTPHeaders,
    IMutableHTTPHeaders,
    MutableHTTPHeaders,
    RawHeaders,
    getFromRawHeaders,
    headerNameAsBytes,
    headerNameAsText,
    headerValueAsBytes,
    headerValueAsText,
    normalizeHeaderName,
    normalizeRawHeaders,
    normalizeRawHeadersFrozen,
)
from ._trial import TestCase
from .not_hypothesis import (
    binary,
    bytesHeaderPairs,
    given,
    latin1_text,
    text,
    textHeaderPairs,
)


__all__ = ()


T = TypeVar("T")
DrawCallable = Callable[[Callable[..., T]], T]


def encodeName(name: str) -> Optional[bytes]:
    return name.encode(HEADER_NAME_ENCODING)


def encodeValue(name: str) -> Optional[bytes]:
    return name.encode(HEADER_VALUE_ENCODING)


def decodeName(name: bytes) -> str:
    return name.decode(HEADER_NAME_ENCODING)


def decodeValue(name: bytes) -> str:
    return name.decode(HEADER_VALUE_ENCODING)


def headerValueSanitize(value: AnyStr) -> AnyStr:
    """
    Sanitize a header value by replacing linear whitespace with spaces.
    """
    if isinstance(value, bytes):
        lws = [b"\r\n", b"\r", b"\n"]
        space = b" "
    else:
        lws = ["\r\n", "\r", "\n"]
        space = " "
    for lw in lws:
        value = value.replace(lw, space)
    return value


class EncodingTests(TestCase):
    """
    Tests for encoding support in L{klein._headers}.
    """

    @given(binary())
    def test_headerNameAsBytesWithBytes(self, name: bytes) -> None:
        """
        L{headerNameAsBytes} passes through L{bytes}.
        """
        self.assertIdentical(headerNameAsBytes(name), name)

    @given(latin1_text(min_size=1))
    def test_headerNameAsBytesWithText(self, name: str) -> None:
        """
        L{headerNameAsBytes} encodes L{str} using L{HEADER_NAME_ENCODING}.
        """
        rawName = encodeName(name)
        self.assertEqual(headerNameAsBytes(name), rawName)

    @given(binary())
    def test_headerNameAsTextWithBytes(self, name: bytes) -> None:
        """
        L{headerNameAsText} decodes L{bytes} using L{HEADER_NAME_ENCODING}.
        """
        self.assertEqual(headerNameAsText(name), decodeName(name))

    @given(text(min_size=1))
    def test_headerNameAsTextWithText(self, name: str) -> None:
        """
        L{headerNameAsText} passes through L{str}.
        """
        self.assertIdentical(headerNameAsText(name), name)

    @given(binary())
    def test_headerValueAsBytesWithBytes(self, value: bytes) -> None:
        """
        L{headerValueAsBytes} passes through L{bytes}.
        """
        self.assertIdentical(headerValueAsBytes(value), value)

    @given(latin1_text(min_size=1))
    def test_headerValueAsBytesWithText(self, value: str) -> None:
        """
        L{headerValueAsBytes} encodes L{str} using L{HEADER_VALUE_ENCODING}.
        """
        rawValue = encodeValue(value)
        self.assertEqual(headerValueAsBytes(value), rawValue)

    @given(binary())
    def test_headerValueAsTextWithBytes(self, value: bytes) -> None:
        """
        L{headerValueAsText} decodes L{bytes} using L{HEADER_VALUE_ENCODING}.
        """
        self.assertEqual(headerValueAsText(value), decodeValue(value))

    @given(text(min_size=1))
    def test_headerValueAsTextWithText(self, value: str) -> None:
        """
        L{headerValueAsText} passes through L{str}.
        """
        self.assertIdentical(headerValueAsText(value), value)


class HeaderNameNormalizationTests(TestCase):
    """
    Tests for header name normalization.
    """

    def test_normalizeLowerCase(self) -> None:
        """
        L{normalizeHeaderName} normalizes header names to lower case.
        """
        self.assertEqual(normalizeHeaderName("FooBar"), "foobar")


class RawHeadersConversionTests(TestCase):
    """
    Tests for L{normalizeRawHeaders}.
    """

    def test_pairWrongLength(self) -> None:
        """
        L{normalizeRawHeaders} raises L{ValueError} if the C{headerPairs}
        argument is not an iterable of 2-item iterables.
        """
        for invalidPair in ((b"k",), (b"k", b"v", b"x")):
            e = self.assertRaises(
                ValueError,
                tuple,
                normalizeRawHeaders(
                    cast(Iterable[Iterable[bytes]], (invalidPair,))
                ),
            )
            self.assertEqual(str(e), "header pair must be a 2-item iterable")

    @given(latin1_text())
    def test_pairNameText(self, name: str) -> None:
        """
        L{normalizeRawHeaders} converts ISO-8859-1-encodable text names into
        bytes.
        """
        rawHeaders = ((name, b"value"),)
        normalized = tuple(normalizeRawHeaders(rawHeaders))

        self.assertEqual(
            normalized,
            ((normalizeHeaderName(headerNameAsBytes(name)), b"value"),),
        )

    @given(latin1_text())
    def test_pairValueText(self, value: str) -> None:
        """
        L{normalizeRawHeaders} converts ISO-8859-1-encodable text values into
        bytes.
        """
        rawHeaders = ((b"name", value),)
        normalized = tuple(normalizeRawHeaders(rawHeaders))

        self.assertEqual(normalized, ((b"name", headerValueAsBytes(value)),))


class GetValuesTestsMixIn(ABC):
    """
    Tests for utilities that access data from the C{RawHeaders} internal
    representation.
    """

    @abstractmethod
    def getValues(
        self, rawHeaders: RawHeaders, name: AnyStr
    ) -> Iterable[AnyStr]:
        """
        Look up the values for the given header name from the given raw
        headers.

        This is called by the other tests in this mix-in class to allow test
        cases that use it to specify how to perform this look-up in the
        implementation being tested.
        """

    def test_getBytesName(self) -> None:
        """
        C{getValues} returns an iterable of L{bytes} values for the
        given L{bytes} header name.
        """
        rawHeaders = ((b"a", b"1"), (b"b", b"2"), (b"c", b"3"), (b"B", b"TWO"))

        normalized: Dict[bytes, List[bytes]] = defaultdict(list)
        for name, value in rawHeaders:
            normalized[normalizeHeaderName(name)].append(value)

        for name, values in normalized.items():
            cast(TestCase, self).assertEqual(
                list(self.getValues(rawHeaders, name)),
                values,
                f"header name: {name!r}",
            )

    def headerNormalize(self, value: str) -> str:
        """
        Test hook for the normalization of header text values, which is a
        behavior Twisted has changed after version 18.9.0.
        """
        return value

    @given(textHeaderPairs())
    def test_getTextName(self, textPairs: Iterable[Tuple[str, str]]) -> None:
        """
        C{getValues} returns an iterable of L{str} values for
        the given L{str} header name.

        This test only inserts Latin1 text into the header values, which is
        valid data.
        """
        textHeaders = tuple(
            (name, headerValueSanitize(value)) for name, value in textPairs
        )

        textValues: Dict[str, List[str]] = defaultdict(list)
        for name, value in textHeaders:
            textValues[normalizeHeaderName(name)].append(value)

        rawHeaders = tuple(
            (headerNameAsBytes(name), headerValueAsBytes(value))
            for name, value in textHeaders
        )

        for name, _values in textValues.items():
            cast(TestCase, self).assertEqual(
                list(self.getValues(rawHeaders, name)),
                [self.headerNormalize(value) for value in _values],
                f"header name: {name!r}",
            )

    @given(bytesHeaderPairs())
    def test_getTextNameBinaryValues(
        self, pairs: Iterable[Tuple[str, bytes]]
    ) -> None:
        """
        C{getValues} returns an iterable of L{str} values for
        the given L{str} header name.

        This test only inserts binary data into the header values, which
        includes invalid data if you are a sane person, but arguably
        technically valid if you read the spec because the spec is unclear
        about header encodings, so we made sure that works also, if only sort
        of.
        """
        rawHeaders = tuple(
            (headerNameAsBytes(name), headerValueSanitize(value))
            for name, value in pairs
        )

        binaryValues: Dict[str, List[bytes]] = defaultdict(list)
        for name, value in rawHeaders:
            binaryValues[headerNameAsText(normalizeHeaderName(name))].append(
                value
            )

        for textName, values in binaryValues.items():
            cast(TestCase, self).assertEqual(
                tuple(self.getValues(rawHeaders, textName)),
                tuple(
                    self.headerNormalize(headerValueAsText(value))
                    for value in values
                ),
                f"header name: {textName!r}",
            )

    def test_getInvalidNameType(self) -> None:
        """
        C{getValues} raises L{TypeError} when the given header name is of an
        unknown type.
        """
        e = cast(TestCase, self).assertRaises(
            TypeError, self.getValues, (), object()
        )
        cast(TestCase, self).assertRegex(
            str(e), "name <object object at 0x[0-9a-f]+> must be str or bytes"
        )


class RawHeadersReadTests(GetValuesTestsMixIn, TestCase):
    """
    Tests for utilities that access data from the "headers tartare" internal
    representation.
    """

    def getValues(
        self, rawHeaders: RawHeaders, name: AnyStr
    ) -> Iterable[AnyStr]:
        return getFromRawHeaders(normalizeRawHeadersFrozen(rawHeaders), name)


class FrozenHTTPHeadersTests(GetValuesTestsMixIn, TestCase):
    """
    Tests for L{FrozenHTTPHeaders}.
    """

    def getValues(
        self, rawHeaders: RawHeaders, name: AnyStr
    ) -> Iterable[AnyStr]:
        headers = FrozenHTTPHeaders(rawHeaders=rawHeaders)
        return headers.getValues(name=name)

    def test_interface(self) -> None:
        """
        L{FrozenHTTPHeaders} implements L{IHTTPHeaders}.
        """
        headers = FrozenHTTPHeaders(rawHeaders=())
        self.assertProvides(IHTTPHeaders, headers)

    def test_defaultHeaders(self) -> None:
        """
        L{FrozenHTTPHeaders.rawHeaders} is empty by default.
        """
        headers = FrozenHTTPHeaders()
        self.assertEqual(headers.rawHeaders, ())


class MutableHTTPHeadersTestsMixIn(GetValuesTestsMixIn, ABC):
    """
    Tests for L{IMutableHTTPHeaders} implementations.
    """

    def assertRawHeadersEqual(
        self, rawHeaders1: RawHeaders, rawHeaders2: RawHeaders
    ) -> None:
        cast(TestCase, self).assertEqual(rawHeaders1, rawHeaders2)

    @abstractmethod
    def headers(self, rawHeaders: RawHeaders) -> IMutableHTTPHeaders:
        """
        Given a L{RawHeaders}, return an L{IMutableHTTPHeaders}.
        """

    def getValues(
        self, rawHeaders: RawHeaders, name: AnyStr
    ) -> Iterable[AnyStr]:
        headers = self.headers(rawHeaders=rawHeaders)
        return headers.getValues(name=name)

    def test_interface(self) -> None:
        """
        Class implements L{IMutableHTTPHeaders}.
        """
        headers = self.headers(rawHeaders=())
        cast(TestCase, self).assertProvides(IMutableHTTPHeaders, headers)

    def test_rawHeaders(self) -> None:
        """
        L{IMutableHTTPHeaders.rawHeaders} equals the raw headers passed at init
        time as a tuple.
        """
        rawHeaders = [(b"a", b"1"), (b"b", b"2"), (b"c", b"3")]
        headers = self.headers(rawHeaders=rawHeaders)
        self.assertRawHeadersEqual(headers.rawHeaders, tuple(rawHeaders))

    def test_removeBytesName(self) -> None:
        """
        L{IMutableHTTPHeaders.remove} removes all values for the given L{bytes}
        header name.
        """
        rawHeaders = ((b"a", b"1"), (b"b", b"2a"), (b"c", b"3"), (b"b", b"2b"))
        headers = self.headers(rawHeaders=rawHeaders)
        headers.remove(name=b"b")

        self.assertRawHeadersEqual(
            headers.rawHeaders, ((b"a", b"1"), (b"c", b"3"))
        )

    def test_removeTextName(self) -> None:
        """
        L{IMutableHTTPHeaders.remove} removes all values for the given L{str}
        header name.
        """
        rawHeaders = ((b"a", b"1"), (b"b", b"2a"), (b"c", b"3"), (b"b", b"2b"))
        headers = self.headers(rawHeaders=rawHeaders)
        headers.remove(name="b")

        self.assertRawHeadersEqual(
            headers.rawHeaders, ((b"a", b"1"), (b"c", b"3"))
        )

    def test_removeInvalidNameType(self) -> None:
        """
        L{IMutableHTTPHeaders.remove} raises L{TypeError} when the given header
        name is of an unknown type.
        """
        headers = self.headers(rawHeaders=())

        e = cast(TestCase, self).assertRaises(
            TypeError, headers.remove, object()
        )
        cast(TestCase, self).assertRegex(
            str(e), "name <object object at 0x[0-9a-f]+> must be str or bytes"
        )

    def test_addValueBytesName(self) -> None:
        """
        L{IMutableHTTPHeaders.addValue} adds the given L{bytes} value for the
        given L{bytes} header name.
        """
        rawHeaders = ((b"a", b"1"), (b"b", b"2a"))
        headers = self.headers(rawHeaders=rawHeaders)
        headers.addValue(name=b"b", value=b"2b")

        self.assertRawHeadersEqual(
            headers.rawHeaders, ((b"a", b"1"), (b"b", b"2a"), (b"b", b"2b"))
        )

    def test_addValueTextName(self) -> None:
        """
        L{IMutableHTTPHeaders.addValue} adds the given L{str} value for the
        given L{str} header name.
        """
        rawHeaders = ((b"a", b"1"), (b"b", b"2a"))
        headers = self.headers(rawHeaders=rawHeaders)
        headers.addValue(name="b", value="2b")

        self.assertRawHeadersEqual(
            headers.rawHeaders, ((b"a", b"1"), (b"b", b"2a"), (b"b", b"2b"))
        )

    def test_addValueBytesNameTextValue(self) -> None:
        """
        L{IMutableHTTPHeaders.addValue} raises L{TypeError} when the given
        header name is L{bytes} and the given value is L{str}.
        """
        headers = self.headers(rawHeaders=())

        e = cast(TestCase, self).assertRaises(
            TypeError, headers.addValue, b"a", "1"
        )
        cast(TestCase, self).assertRegex(
            str(e), "value u?'1' must be bytes to match name b?'a'"
        )

    def test_addValueTextNameBytesValue(self) -> None:
        """
        L{IMutableHTTPHeaders.addValue} raises L{TypeError} when the given
        header name is L{str} and the given value is L{bytes}.
        """
        headers = self.headers(rawHeaders=())

        e = cast(TestCase, self).assertRaises(
            TypeError, headers.addValue, "a", b"1"
        )
        cast(TestCase, self).assertRegex(
            str(e), "value b?'1' must be str to match name u?'a'"
        )

    def test_addValueInvalidNameType(self) -> None:
        """
        L{IMutableHTTPHeaders.addValue} raises L{TypeError} when the given
        header name is of an unknown type.
        """
        headers = self.headers(rawHeaders=())

        e = cast(TestCase, self).assertRaises(
            TypeError, headers.addValue, object(), b"1"
        )
        cast(TestCase, self).assertRegex(
            str(e), "name <object object at 0x[0-9a-f]+> must be str or bytes"
        )


class MutableHTTPHeadersTests(MutableHTTPHeadersTestsMixIn, TestCase):
    """
    Tests for L{MutableHTTPHeaders}.
    """

    def headers(self, rawHeaders: RawHeaders) -> IMutableHTTPHeaders:
        return MutableHTTPHeaders(rawHeaders=rawHeaders)

    def test_defaultHeaders(self) -> None:
        """
        L{MutableHTTPHeaders.rawHeaders} is empty by default.
        """
        headers = MutableHTTPHeaders()
        self.assertEqual(headers.rawHeaders, ())



# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein.test._trial}.
"""

from zope.interface import Interface, implementer

from ._trial import TestCase


__all__ = ()


class TestCaseTests(TestCase):
    """
    Tests for L{TestCase}.
    """

    class IFrobbable(Interface):
        """
        Frobbable object.
        """

        def frob() -> None:
            """
            Frob the object.
            """

    @implementer(IFrobbable)
    class Frobbable:
        """
        Implements L{IFrobbable}.
        """

        def frob(self) -> None:
            pass

    @implementer(IFrobbable)
    class NotFrobbable:  # type: ignore[misc]  # …intentional for test
        """
        Does not implement L{IFrobbable}, despite declaring.
        """

    def test_assertProvidesPass(self) -> None:
        """
        L{TestCase.assertProvides} does not raise when C{interface} is provided
        by C{obj}.
        """
        frobbable = self.Frobbable()
        self.assertProvides(self.IFrobbable, frobbable)
        frobbable.frob()  # Coverage

    def test_assertProvidesFail(self) -> None:
        """
        L{TestCase.assertProvides} does not raise when C{interface} is not
        provided by C{obj}.
        """
        self.assertRaises(
            self.failureException,
            self.assertProvides,
            self.IFrobbable,
            self.NotFrobbable(),
        )



# -*- test-case-name: klein.test.test_irequest -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Support for interoperability with L{twisted.web.iweb.IRequest}.
"""

from io import BytesIO
from typing import cast

from attr import Factory, attrib, attrs
from attr.validators import provides
from hyperlink import DecodedURL
from tubes.itube import IFount
from zope.interface import implementer

from twisted.python.compat import nativeString
from twisted.web.iweb import IRequest

from ._headers import IHTTPHeaders
from ._headers_compat import HTTPHeadersWrappingHeaders
from ._message import FountAlreadyAccessedError, MessageState
from ._request import IHTTPRequest
from ._tubes import IOFount, fountToBytes


__all__ = ()


noneIO = BytesIO()


@implementer(IHTTPRequest)
@attrs(frozen=True)
class HTTPRequestWrappingIRequest:
    """
    HTTP request.

    This is an L{IHTTPRequest} implementation that wraps an L{IRequest} object.
    """

    _request: IRequest = attrib(validator=provides(IRequest))

    _state: MessageState = attrib(default=Factory(MessageState), init=False)

    @property
    def method(self) -> str:
        return cast(str, self._request.method.decode("ascii"))

    @property
    def uri(self) -> DecodedURL:
        request = self._request

        # This code borrows from t.w.server.Request._prePathURL.

        if request.isSecure():
            scheme = "https"
        else:
            scheme = "http"

        netloc = nativeString(request.getRequestHostname())

        port = request.getHost().port
        if request.isSecure():
            default = 443
        else:
            default = 80
        if port != default:
            netloc += f":{port}"

        path = nativeString(request.uri)
        if path and path[0] == "/":
            path = path[1:]

        return DecodedURL.fromText(f"{scheme}://{netloc}/{path}")

    @property
    def headers(self) -> IHTTPHeaders:
        return HTTPHeadersWrappingHeaders(headers=self._request.requestHeaders)

    def bodyAsFount(self) -> IFount:
        source = self._request.content
        if source is noneIO:
            raise FountAlreadyAccessedError()

        fount = IOFount(source=source)

        self._request.content = noneIO

        return fount

    async def bodyAsBytes(self) -> bytes:
        if self._state.cachedBody is not None:
            return self._state.cachedBody  # pragma: no cover

        fount = self.bodyAsFount()
        self._state.cachedBody = await fountToBytes(fount)
        return self._state.cachedBody



from ._app import (
    Klein,
    KleinRequest,
    handle_errors,
    resource,
    route,
    run,
    subroute,
    url_for,
    urlFor,
)


__all__ = (
    "Klein",
    "KleinRequest",
    "handle_errors",
    "resource",
    "route",
    "run",
    "subroute",
    "urlFor",
    "url_for",
)



# -*- test-case-name: klein.test.test_request -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
HTTP request API.
"""

from typing import Union

from attr import Factory, attrib, attrs
from attr.validators import instance_of, provides
from hyperlink import DecodedURL
from tubes.itube import IFount
from zope.interface import implementer

from ._imessage import IHTTPHeaders, IHTTPRequest
from ._message import MessageState, bodyAsBytes, bodyAsFount, validateBody


__all__ = ()


@implementer(IHTTPRequest)
@attrs(frozen=True)
class FrozenHTTPRequest:
    """
    Immutable HTTP request.
    """

    method: str = attrib(validator=instance_of(str))
    uri: DecodedURL = attrib(validator=instance_of(DecodedURL))
    headers: IHTTPHeaders = attrib(validator=provides(IHTTPHeaders))

    _body: Union[bytes, IFount] = attrib(validator=validateBody)

    _state: MessageState = attrib(default=Factory(MessageState), init=False)

    def bodyAsFount(self) -> IFount:
        return bodyAsFount(self._body, self._state)

    async def bodyAsBytes(self) -> bytes:
        return await bodyAsBytes(self._body, self._state)



# Copyright (c) 2011-2021. See LICENSE for details.

"""
Interfaces related to HTTP messages.

Do not import directly from here, except:
 - From _interfaces.py.
 - From implementations of these interfaces, but even then, import the
   zope.interface.Interface classes via _interfaces.py.

This will ensure that type checking works.
"""

from typing import AnyStr, Iterable, MutableSequence, Sequence, Tuple

from hyperlink import DecodedURL
from tubes.itube import IFount
from zope.interface import Attribute, Interface


__all__ = ()


RawHeader = Tuple[bytes, bytes]
RawHeaders = Sequence[RawHeader]
MutableRawHeaders = MutableSequence[RawHeader]


class FountAlreadyAccessedError(Exception):
    """
    The HTTP message's fount has already been accessed and is no longer
    available.
    """


class IHTTPHeaders(Interface):
    """
    HTTP entity headers.

    HTTP headers names and values are sort-of-but-not-quite-specifically
    expected to be text.

    Because the specifications are somewhat vague, and implementations vary in
    fidelity, both header names and values must be made available as the
    original bytes received from the network.
    This interface also makes them available as an ordered sequence of name and
    value pairs so that they can be iterated in the same order as they were
    received on the network.

    As such, the C{rawHeaders} attribute provides the header data as a sequence
    of C{(name, value)} L{tuple}s.

    A dictionary-like interface that maps text names to an ordered sequence of
    text values.
    This interface assumes that both header name bytes and header value bytes
    are encoded as ISO-8859-1.

    Note that header name bytes should be strictly encoded as ASCII; this
    interface uses ISO-8859-1 to provide interoperability with (naughty) HTTP
    implementations that send non-ASCII data.
    Because ISO-8859-1 is a superset of ASCII, this will still work for
    well-behaved implementations.
    """

    rawHeaders: RawHeaders = Attribute(
        """
        Raw header data as a tuple in the from: C{((name, value), ...)}.
        C{name} and C{value} are bytes.
        Headers are provided in the order that they were received.
        Headers with multiple values are provided as separate name and value
        pairs.
        """
    )

    def getValues(name: AnyStr) -> Iterable[AnyStr]:
        """
        Get the values associated with the given header name.

        If the given name is L{bytes}, the value will be returned as the raw
        header L{bytes}.

        If the given name is L{str}, the name will be encoded as ISO-8859-1
        and the value will be returned as text, by decoding the raw header
        value bytes with ISO-8859-1.

        @param name: The name of the header to look for.

        @return: The values of the header with the given name.
        """


class IMutableHTTPHeaders(IHTTPHeaders):
    """
    Mutable HTTP entity headers.
    """

    def remove(name: AnyStr) -> None:
        """
        Remove all header name/value pairs for the given header name.

        If the given name is L{str}, it will be encoded as ISO-8859-1 before
        comparing to the (L{bytes}) header names.

        @param name: The name of the header to remove.
        """

    def addValue(name: AnyStr, value: AnyStr) -> None:
        """
        Add the given header name/value pair.

        If the given name is L{bytes}, the value must also be L{bytes}.

        If the given name is L{str}, it will be encoded as ISO-8859-1, and the
        value, which must also be L{str}, will be encoded as ISO-8859-1.
        """


class IHTTPMessage(Interface):
    """
    HTTP entity.
    """

    headers: IHTTPHeaders = Attribute("Entity headers.")

    def bodyAsFount() -> IFount:
        """
        The entity body, as a fount.

        @note: The fount may only be accessed once.
            It provides a mechanism for accessing the body as a stream of data,
            potentially as it is read from the network, without having to cache
            the entire body, which may be large.
            Because there is no caching, it is not possible to "start over" by
            accessing the fount a second time.
            Attempting to do so will raise L{FountAlreadyAccessedError}.

        @raise FountAlreadyAccessedError: If the fount has previously been
            accessed.
        """

    async def bodyAsBytes() -> bytes:
        """
        The entity body, as bytes.

        @note: This necessarily reads the entire entity body into memory,
            which may be a problem if the body is large.

        @note: This method caches the body, which means that unlike
            C{self.bodyAsFount}, calling it repeatedly will return the same
            data.

        @note: This method accesses the fount (via C{self.bodyAsFount}), which
            means the fount will not be available afterwards, and that if
            C{self.bodyAsFount} has previously been called directly, this
            method will raise L{FountAlreadyAccessedError}.

        @raise FountAlreadyAccessedError: If the fount has previously been
            accessed.
        """


class IHTTPRequest(IHTTPMessage):
    """
    HTTP request.
    """

    method: str = Attribute("Request method.")
    uri: DecodedURL = Attribute("Request URI.")


class IHTTPResponse(IHTTPMessage):
    """
    HTTP response.
    """

    status: int = Attribute("Response status code.")



from typing import Any, List, Tuple, cast
from xml.etree import ElementTree

import attr
from treq import content
from treq.testing import StubTreq

from twisted.internet.defer import inlineCallbacks
from twisted.python.compat import nativeString
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.error import FlattenerError
from twisted.web.iweb import IRequest
from twisted.web.template import Element, TagLoader, renderer, tags

from klein import (
    Field,
    Form,
    Klein,
    KleinRenderable,
    RenderableForm,
    Requirer,
    SessionProcurer,
)
from klein.interfaces import (
    ISession,
    ISessionStore,
    NoSuchSession,
    SessionMechanism,
    ValidationError,
)
from klein.storage.memory import MemorySessionStore

from .._form import textConverter


class DanglingField(Field):
    """
    A dangling field that, for some reason, doesn't remember its own name when
    told.
    """

    def maybeNamed(self, name: str) -> Field:
        return self


@attr.s(auto_attribs=True, hash=False)
class TestObject:
    sessionStore: ISessionStore
    calls: List = attr.ib(factory=list)

    router = Klein()
    requirer = Requirer()

    @requirer.prerequisite([ISession])
    @inlineCallbacks
    def procureASession(self, request: IRequest) -> Any:
        try:
            yield (
                SessionProcurer(
                    self.sessionStore, secureTokenHeader=b"X-Test-Session"
                ).procureSession(request)
            )
        except NoSuchSession:
            # Intentionally slightly buggy - if a session can't be procured,
            # simply leave it out and rely on checkCSRF to ensure the session
            # component is present before proceeding.
            pass

    @requirer.require(
        router.route("/dangling-param", methods=["POST"]),
        dangling=DanglingField(lambda x: x, "text"),
    )
    def danglingParameter(self, dangling: str) -> None:
        """
        Provided to test error reporting on handling of dangling fields.

        @see: L{DanglingField}
        """

    @requirer.require(
        router.route("/dangling-param", methods=["GET"]),
        form=Form.rendererFor(danglingParameter, action="/dangling-param"),
    )
    def renderDanglingParameter(self, form: Form) -> Form:
        """
        Provided to test error reporting on rendering of dangling fields.

        @see: L{DanglingField}
        """
        return form

    @requirer.require(
        router.route("/handle", methods=["POST"]),
        name=Field.text(),
        value=Field.number(),
    )
    def handler(self, name: str, value: float) -> bytes:
        self.calls.append((name, value))
        return b"yay"

    @requirer.require(
        router.route("/handle-submit", methods=["POST"]),
        name=Field.text(),
        button=Field.submit("OK"),
    )
    def handlerWithSubmit(self, name: str, button: str) -> None:
        """
        Form with a submit button.
        """

    @requirer.require(
        router.route("/password-field", methods=["POST"]), pw=Field.password()
    )
    def gotPassword(self, pw: str) -> bytes:
        self.calls.append(("password", pw))
        return b"password received"

    @requirer.require(
        router.route("/notrequired", methods=["POST"]),
        name=Field.text(),
        value=Field.number(required=False, default=7.0),
    )
    def notRequired(self, name: str, value: float) -> bytes:
        self.calls.append((name, value))
        return b"okay"

    @requirer.require(
        router.route("/constrained", methods=["POST"]),
        goldilocks=Field.number(minimum=3, maximum=9),
    )
    def constrained(self, goldilocks: int) -> bytes:
        self.calls.append(("constrained", goldilocks))
        return b"got it"

    @requirer.require(
        router.route("/render", methods=["GET"]),
        form=Form.rendererFor(handler, action="/handle"),
    )
    def renderer(self, form: Form) -> Form:
        return form

    @requirer.require(
        router.route("/render-submit", methods=["GET"]),
        form=Form.rendererFor(handlerWithSubmit, action="/handle-submit"),
    )
    def submitRenderer(self, form: RenderableForm) -> RenderableForm:
        return form

    @requirer.require(
        router.route("/render-custom", methods=["GET"]),
        form=Form.rendererFor(handler, action="/handle"),
    )
    def customFormRender(self, form: RenderableForm) -> Any:
        """
        Include just the glue necessary for CSRF protection and let the
        application render the rest of the form.
        """
        return Element(loader=TagLoader(tags.html(tags.body(form.glue()))))

    @requirer.require(
        router.route("/render-cascade", methods=["GET"]),
        form=Form.rendererFor(handler, action="/handle"),
    )
    def cascadeRenderer(self, form: RenderableForm) -> Element:
        class CustomElement(Element):
            @renderer
            def customize(self, request: IRequest, tag: Any) -> Any:
                return tag("customized")

        form.validationErrors[form._form.fields[0]] = ValidationError(
            message=(tags.div(class_="checkme", render="customize"))
        )

        return CustomElement(loader=TagLoader(form))

    @requirer.require(
        router.route("/handle-validation", methods=["POST"]),
        value=Field.number(maximum=10),
    )
    def customValidation(self, value: int) -> None:
        """
        never called.
        """

    @requirer.require(Form.onValidationFailureFor(customValidation))
    def customFailureHandling(self, values: RenderableForm) -> bytes:
        """
        Handle validation failure.
        """
        self.calls.append(("validation", values))
        return b"~special~"

    @requirer.require(
        router.route("/handle-empty", methods=["POST"]),
    )
    def emptyHandler(self) -> KleinRenderable:
        """
        Empty form handler; just for testing rendering.
        """

    @requirer.require(
        router.route("/render-empty", methods=["GET"]),
        form=Form.rendererFor(emptyHandler, action="/handle-empty"),
    )
    def emptyRenderer(self, form: RenderableForm) -> RenderableForm:
        return form


def simpleFormRouter() -> Tuple[Klein, List[Tuple[str, int]]]:
    """
    Create a simple router hooked up to a field handler.
    """
    router = Klein()
    requirer = Requirer()
    calls = []

    @requirer.require(
        router.route("/getme", methods=["GET"]),
        name=Field.text(),
        value=Field.number(),
        custom=Field(formInputType="number", converter=int, required=False),
    )
    def justGet(name: str, value: int, custom: int) -> bytes:
        calls.append((name, value or custom))
        return b"got"

    return router, calls


class TestForms(SynchronousTestCase):
    """
    Tests for L{klein.Form} and associated tools.
    """

    def test_textConverter(self) -> None:
        """
        Convert a string of either type to text.
        """
        text = "f\xf6o"
        for string in (text, text.encode("utf-8")):
            result = textConverter(string)  # type: ignore[type-var]
            self.assertIsInstance(result, str)
            self.assertEqual(result, text)

    def test_handling(self) -> None:
        """
        A handler for a Form with Fields receives those fields as input, as
        passed by an HTTP client.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                data=dict(name="hello", value="1234", ignoreme="extraneous"),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(self.successResultOf(content(response)), b"yay")
        self.assertEqual(to.calls, [("hello", 1234)])

    def test_handlingPassword(self) -> None:
        """
        From the perspective of form handling, passwords are handled like
        strings.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/password-field",
                data=dict(pw="asdfjkl;"),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(
            self.successResultOf(content(response)), b"password received"
        )
        self.assertEqual(to.calls, [("password", "asdfjkl;")])

    def test_numberConstraints(self) -> None:
        """
        Number parameters have minimum and maximum validations and the object
        will not be called when the values exceed them.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        tooLow = self.successResultOf(
            stub.post(
                "https://localhost/constrained",
                data=dict(goldilocks="1"),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        tooHigh = self.successResultOf(
            stub.post(
                "https://localhost/constrained",
                data=dict(goldilocks="20"),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        justRight = self.successResultOf(
            stub.post(
                "https://localhost/constrained",
                data=dict(goldilocks="7"),
                headers={b"X-Test-Session": session.identifier},
            )
        )

        self.assertEqual(tooHigh.code, 400)
        self.assertEqual(tooLow.code, 400)
        self.assertEqual(justRight.code, 200)
        self.assertEqual(self.successResultOf(content(justRight)), b"got it")
        self.assertEqual(to.calls, [("constrained", 7)])

    def test_missingRequiredParameter(self) -> None:
        """
        If required fields are missing, a default error form is presented and
        the form's handler is not called.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                data=dict(),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 400)
        self.assertIn(
            b"a value was required but none was supplied",
            self.successResultOf(content(response)),
        )
        self.assertEqual(to.calls, [])

    def test_noName(self) -> None:
        """
        A handler for a Form with a Field that doesn't have a name will return
        an error explaining the problem when either processed in a handler or
        rendered in a form.

        @see: L{DanglingField}.
        """
        mem = MemorySessionStore()
        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )
        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/dangling-param",
                data=dict(),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 500)
        errors = self.flushLoggedErrors(ValueError)
        self.assertEqual(len(errors), 1)
        self.assertIn(
            "Cannot extract unnamed form field.",
            str(errors[0].value),
        )

        self.successResultOf(
            stub.get(
                "https://localhost/dangling-param",
                headers={b"X-Test-Session": session.identifier},
            ),
        )
        self.assertEqual(response.code, 500)
        errors = self.flushLoggedErrors(FlattenerError)
        self.assertEqual(len(errors), 1)
        self.assertIn(
            "Cannot generate tags for unnamed form field.",
            str(errors[0].value.args[0]),
        )

    def test_handlingGET(self) -> None:
        """
        A GET handler for a Form with Fields receives query parameters matching
        those field names as input.
        """
        router, calls = simpleFormRouter()

        stub = StubTreq(router.resource())

        response = self.successResultOf(
            stub.get(
                b"https://localhost/getme?name=hello,%20big+world&value=4321"
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(self.successResultOf(content(response)), b"got")
        self.assertEqual(calls, [("hello, big world", 4321)])

    def test_validatingParameters(self) -> None:
        """
        When a parameter fails to validate - for example, a non-number passed
        to a numeric Field, the request fails with a 400 and the default
        validation failure handler displays a form which explains the error.
        """
        router, calls = simpleFormRouter()

        stub = StubTreq(router.resource())

        response = self.successResultOf(
            stub.get(
                b"https://localhost/getme?"
                b"name=hello,%20big+world&value=not+a+number"
            )
        )
        responseForm = self.successResultOf(content(response))
        self.assertEqual(response.code, 400)
        self.assertEqual(calls, [])
        responseForm = self.successResultOf(content(response))
        responseDom = ElementTree.fromstring(responseForm)
        errors = responseDom.findall(
            ".//*[@class='klein-form-validation-error']"
        )
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].text, "not a valid number")

    def test_customParameterValidation(self) -> None:
        """
        When a custom parameter fails to validate by raising ValueError - for
        example, a non-number passed to a numeric Field, the request fails with
        a 400 and the default validation failure handler displays a form which
        explains the error.
        """
        router, calls = simpleFormRouter()

        stub = StubTreq(router.resource())

        response = self.successResultOf(
            stub.get(
                b"https://localhost/getme?"
                b"name=hello,%20big+world&value=0&custom=not+a+number"
            )
        )
        responseForm = self.successResultOf(content(response))
        self.assertEqual(response.code, 400)
        self.assertEqual(calls, [])
        responseForm = self.successResultOf(content(response))
        responseDom = ElementTree.fromstring(responseForm)
        errors = responseDom.findall(
            ".//*[@class='klein-form-validation-error']"
        )
        self.assertEqual(len(errors), 1)
        errorText = cast(str, errors[0].text)
        self.assertIsNot(errorText, None)
        self.assertEqual(
            errorText,
            "invalid literal for int() with base 10: 'not a number'",
        )

    def test_handlingJSON(self) -> None:
        """
        A handler for a form with Fields receives those fields as input, as
        passed by an HTTP client that submits a JSON POST body.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                json=dict(name="hello", value="1234", ignoreme="extraneous"),
                headers={"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(self.successResultOf(content(response)), b"yay")
        self.assertEqual(to.calls, [("hello", 1234)])

    def test_missingOptionalParameterJSON(self) -> None:
        """
        If a required Field is missing from the JSON body, its default value is
        used.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/notrequired",
                json=dict(name="one"),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        response2 = self.successResultOf(
            stub.post(
                "https://localhost/notrequired",
                json=dict(name="two", value=2),
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(response2.code, 200)
        self.assertEqual(self.successResultOf(content(response)), b"okay")
        self.assertEqual(self.successResultOf(content(response2)), b"okay")
        self.assertEqual(to.calls, [("one", 7.0), ("two", 2.0)])

    def test_rendering(self) -> None:
        """
        When a route requires form fields, it renders a form with those fields.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(
            stub.get(
                "https://localhost/render",
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertIn(
            response.headers.getRawHeaders(b"content-type")[0], b"text/html"
        )
        responseDom = ElementTree.fromstring(
            self.successResultOf(content(response))
        )
        submitButton = responseDom.findall(".//*[@type='submit']")
        self.assertEqual(len(submitButton), 1)
        self.assertEqual(
            submitButton[0].attrib["name"], "__klein_auto_submit__"
        )

    def test_renderingExplicitSubmit(self) -> None:
        """
        When a form renderer specifies a submit button, no automatic submit
        button is rendered.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(
            stub.get(
                "https://localhost/render-submit",
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertIn(
            response.headers.getRawHeaders(b"content-type")[0], b"text/html"
        )
        responseDom = ElementTree.fromstring(
            self.successResultOf(content(response))
        )
        submitButton = responseDom.findall(".//*[@type='submit']")
        self.assertEqual(len(submitButton), 1)
        self.assertEqual(submitButton[0].attrib["name"], "button")

    def test_renderingFormGlue(self) -> None:
        """
        When a form renderer renders just the glue, none of the rest of the
        form is included.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(
            stub.get(
                "https://localhost/render-custom",
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertIn(
            response.headers.getRawHeaders(b"content-type")[0], b"text/html"
        )
        responseDom = ElementTree.fromstring(
            self.successResultOf(content(response))
        )
        submitButton = responseDom.findall(".//*[@type='submit']")
        self.assertEqual(len(submitButton), 0)
        protectionField = responseDom.findall(
            ".//*[@name='__csrf_protection__']"
        )
        self.assertEqual(protectionField[0].attrib["value"], session.identifier)

    def test_renderingEmptyForm(self) -> None:
        """
        When a form renderer specifies a submit button, no automatic submit
        button is rendered.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(
            stub.get(
                "https://localhost/render-empty",
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertIn(
            response.headers.getRawHeaders(b"content-type")[0], b"text/html"
        )
        responseDom = ElementTree.fromstring(
            self.successResultOf(content(response))
        )
        submitButton = responseDom.findall(".//*[@type='submit']")
        self.assertEqual(len(submitButton), 1)
        self.assertEqual(
            submitButton[0].attrib["name"], "__klein_auto_submit__"
        )
        protectionField = responseDom.findall(
            ".//*[@name='__csrf_protection__']"
        )
        self.assertEqual(protectionField[0].attrib["value"], session.identifier)

    def test_renderLookupError(self) -> None:
        """
        RenderableForm raises L{MissingRenderMethod} if anything attempts to
        look up a render method on it.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(
            stub.get(
                "https://localhost/render-cascade",
                headers={b"X-Test-Session": session.identifier},
            )
        )
        self.assertEqual(response.code, 200)
        # print(self.successResultOf(response.content()).decode('utf-8'))
        failures = self.flushLoggedErrors()
        self.assertEqual(len(failures), 1)
        self.assertIn("MissingRenderMethod", str(failures[0]))

    def test_customValidationHandling(self) -> None:
        """
        L{Form.onValidationFailureFor} handles form validation failures by
        handing its thing a renderable form.
        """
        mem = MemorySessionStore()

        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Header)
        )

        testobj = TestObject(mem)
        stub = StubTreq(testobj.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle-validation",
                headers={b"X-Test-Session": session.identifier},
                json={"value": 300},
            )
        )
        self.assertEqual(response.code, 200)
        self.assertIn(
            response.headers.getRawHeaders(b"content-type")[0], b"text/html"
        )
        responseText = self.successResultOf(content(response))
        self.assertEqual(responseText, b"~special~")
        self.assertEqual(
            [
                (k.pythonArgumentName, v)
                for k, v in testobj.calls[-1][1].prevalidationValues.items()
            ],
            [("value", 300)],
        )

    def test_renderingWithNoSessionYet(self) -> None:
        """
        When a route is rendered with no session, it sets a cookie to establish
        a new session.
        """
        mem = MemorySessionStore()
        stub = StubTreq(TestObject(mem).router.resource())
        response = self.successResultOf(stub.get("https://localhost/render"))
        self.assertEqual(response.code, 200)
        setCookie = response.cookies()["Klein-Secure-Session"]
        expected = f'value="{setCookie}"'
        actual = self.successResultOf(content(response)).decode("utf-8")
        self.assertIn(expected, actual)

    def test_noSessionPOST(self) -> None:
        """
        An unauthenticated, CSRF-protected form will return a 403 Forbidden
        status code.
        """
        mem = MemorySessionStore()
        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                data=dict(name="hello", value="1234"),
            )
        )
        self.assertEqual(to.calls, [])
        self.assertEqual(response.code, 403)
        self.assertIn(b"CSRF", self.successResultOf(content(response)))

    def test_cookieNoToken(self) -> None:
        """
        A cookie-authenticated, CSRF-protected form will return a 403 Forbidden
        status code when a CSRF protection token is not supplied.
        """
        mem = MemorySessionStore()
        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Cookie)
        )
        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                data=dict(name="hello", value="1234", ignoreme="extraneous"),
                cookies={
                    "Klein-Secure-Session": nativeString(session.identifier)
                },
            )
        )
        self.assertEqual(to.calls, [])
        self.assertEqual(response.code, 403)
        self.assertIn(b"CSRF", self.successResultOf(content(response)))

    def test_cookieWithToken(self) -> None:
        """
        A cookie-authenticated, CRSF-protected form will call the form as
        expected.
        """
        mem = MemorySessionStore()
        session = self.successResultOf(
            mem.newSession(True, SessionMechanism.Cookie)
        )
        to = TestObject(mem)
        stub = StubTreq(to.router.resource())
        response = self.successResultOf(
            stub.post(
                "https://localhost/handle",
                data=dict(
                    name="hello",
                    value="1234",
                    ignoreme="extraneous",
                    __csrf_protection__=session.identifier,
                ),
                cookies={
                    "Klein-Secure-Session": nativeString(session.identifier)
                },
            )
        )
        self.assertEqual(to.calls, [("hello", 1234)])
        self.assertEqual(response.code, 200)
        self.assertIn(b"yay", self.successResultOf(content(response)))



# -*- test-case-name: klein.test.test_app -*-
"""
Applications are great.  Lets have more of them.
"""

from __future__ import annotations

import sys
from contextlib import contextmanager
from inspect import iscoroutine
from typing import (
    IO,
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from weakref import ref

from werkzeug.routing import Map, MapAdapter, Rule, Submount
from zope.interface import implementer

from twisted.internet import reactor
from twisted.internet.defer import ensureDeferred
from twisted.internet.endpoints import serverFromString
from twisted.python import log
from twisted.python.components import registerAdapter
from twisted.python.failure import Failure
from twisted.web.iweb import IRenderable, IRequest
from twisted.web.resource import IResource
from twisted.web.server import Request, Site
from twisted.web.template import Tag

from ._decorators import modified, named
from ._interfaces import IKleinRequest, KleinQueryValue
from ._resource import KleinResource, route_metadata
from ._typing_compat import Concatenate, ParamSpec, Protocol


_KleinSynchronousRenderable = Union[
    str, bytes, IResource, IRenderable, Tag, None
]
KleinSynchronousRenderable = Union[
    _KleinSynchronousRenderable, Iterable[_KleinSynchronousRenderable]
]
KleinRenderable = Union[
    KleinSynchronousRenderable, Awaitable[KleinSynchronousRenderable]
]


class KleinErrorFunction(Protocol):
    def __call__(
        _self,
        request: IRequest,
        failure: Failure,
    ) -> KleinRenderable:
        """
        Function that, when registered with L{Klein.handle_errors}, handles
        errors raised during request routing.
        """


class KleinErrorMethod(Protocol):
    def __call__(
        _self,
        self: Any,
        request: IRequest,
        failure: Failure,
    ) -> KleinRenderable:
        """
        Method that, when registered with L{Klein.handle_errors}, handles
        errors raised during request routing.
        """


KleinRouteHandler = Callable[..., KleinRenderable]
"""
A handler for a Klein route.  Sadly it is not possible to tell the type checker
that the first argument is IRequest or the first/second arguments are (Self,
IRequest); this is not a useful actual bound on the type in terms of what you
can do with it, so Mypy can't conceptualize it; the rest of the signature is a
Werkzeug/@require-derived soup.  So "you didn't pass a request" is just an
informative message for the user, not a type assertion; checking the parameter
list will have to be handled at runtime.
"""

KleinRouteHandlerT = TypeVar("KleinRouteHandlerT", bound=KleinRouteHandler)
"""
Let's make sure that we don't modify klein handlers' arg lists as we pass them
through though.
"""

KleinErrorHandler = Union[KleinErrorFunction, KleinErrorMethod]
"""
An error handler for Klein, expressed either as a function or a method.
"""


class RouteMetadata(Protocol):
    """
    Placeholder protocol just enumerating the attributes that are set on route
    functions.
    """

    __name__: str
    segment_count: int


def _call(
    __klein_instance__: Optional[Klein],
    __klein_f__: Callable[..., KleinRenderable],
    *args: Any,
    **kwargs: Any,
) -> KleinRenderable:
    """
    Call C{__klein_f__} with the given C{*args} and C{**kwargs}.

    Insert C{__klein_instance__} as the first positional argument to
    C{__klein_f__} if C{__klein_f__} is not decorated with
    L{klein._decorators.bindable}.

    @return: The result of C{__klein_f__}; additionally, if C{__klein_f__}
        returns a coroutine, instead return the Deferred created by calling
        C{ensureDeferred} on it.
    """
    if __klein_instance__ is not None or getattr(
        __klein_f__, "__klein_bound__", False
    ):
        args = (__klein_instance__,) + args
    result = __klein_f__(*args, **kwargs)
    if iscoroutine(result):
        result = ensureDeferred(result)
    return result


def buildURL(
    mapper: MapAdapter,
    endpoint: str,
    values: Optional[Mapping[str, KleinQueryValue]] = None,
    method: Optional[str] = None,
    force_external: bool = False,
    append_unknown: bool = True,
) -> str:
    return mapper.build(
        endpoint, values, method, force_external, append_unknown
    )


@implementer(IKleinRequest)
class KleinRequest:
    mapper: MapAdapter

    def __init__(self, request: Request) -> None:
        self.branch_segments = [""]

    def url_for(
        self,
        endpoint: str,
        values: Optional[Mapping[str, KleinQueryValue]] = None,
        method: Optional[str] = None,
        force_external: bool = False,
        append_unknown: bool = True,
    ) -> str:
        return buildURL(
            self.mapper,
            endpoint,
            values,
            method,
            force_external,
            append_unknown,
        )


registerAdapter(KleinRequest, Request, IKleinRequest)


ErrorHandlers = List[Tuple[List[Type[Exception]], KleinErrorHandler]]


# begin argument-processing hack to copy all args from current installed
# version of Rule to @route's *args and **kwargs

R = TypeVar("R", covariant=True)
P = ParamSpec("P")


class _PartialRouteSignature(Protocol[R]):
    def __call__(
        _self,
        self: Klein,
        url: str,
        *args: Any,
        branch: bool = False,
        **kwargs: Any,
    ) -> R:
        """
        This is the portion of the signature of C{route} which Klein owns.
        """


class _FullRouteSignature(Protocol[P, R]):
    def __call__(
        _self,
        self: Klein,
        url: str,
        *args: P.args,
        branch: bool = False,
        **kwargs: P.kwargs,
    ) -> R:
        """
        This integrates L{_PartialRouteSignature} with L{Rule}.
        """


def _removeStringArg(rule: Callable[Concatenate[str, P], R]) -> Callable[P, R]:
    """
    We call the 'string' argument to Rule 'url', so we need to remove it.
    """
    return None  # type:ignore[return-value]


def _routeArgsFrom(
    argProvider: Callable[P, object]
) -> Callable[[_PartialRouteSignature[R]], _FullRouteSignature[P, R]]:
    def decorator(
        decoratee: _PartialRouteSignature[R],
    ) -> _FullRouteSignature[P, R]:
        return decoratee

    return decorator


def _normalFunction(arg: Callable[P, R]) -> Callable[P, R]:
    """
    Indicate to Mypy that the decorated callable is in fact a normal
    user-defined function, so that it will be treated as a descriptor with a
    C{self} that binds as a method.

    @see: U{<mypy's confusing treatment of callables>
        https://github.com/python/mypy/issues/15189}
    """
    return arg


# end argument-processing hack


class Klein:
    """
    L{Klein} is an object which is responsible for maintaining the routing
    configuration of our application.

    @ivar _url_map: A C{werkzeug.routing.Map} object which will be used for
        routing resolution.
    @ivar _endpoints: A C{dict} mapping endpoint names to handler functions.
    """

    _subroute_segments = 0

    def __init__(self) -> None:
        self._url_map = Map()
        self._endpoints: Dict[str, KleinRouteHandler] = {}
        self._error_handlers: ErrorHandlers = []
        self._instance: Optional[Klein] = None
        self._boundAs: Optional[str] = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Klein):
            return vars(self) == vars(other)
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    @property
    def url_map(self) -> Map:
        """
        Read only property exposing L{Klein._url_map}.
        """
        return self._url_map

    @property
    def endpoints(self) -> Dict[str, KleinRouteHandler]:
        """
        Read only property exposing L{Klein._endpoints}.
        """
        return self._endpoints

    def execute_endpoint(
        self, endpoint: str, request: IRequest, *args: Any, **kwargs: Any
    ) -> KleinRenderable:
        """
        Execute the named endpoint with all arguments and possibly a bound
        instance.
        """
        endpoint_f: Callable[..., KleinRenderable] = self._endpoints[endpoint]
        # typing note: endpoint_f is a KleinRouteHandler, which is not defined
        # as taking *args, **kwargs (because they aren't required), but we're
        # going to pass them along here anyway.
        return endpoint_f(self._instance, request, *args, **kwargs)

    def execute_error_handler(
        self,
        handler: KleinErrorMethod,
        request: IRequest,
        failure: Failure,
    ) -> KleinRenderable:
        """
        Execute the passed error handler, possibly with a bound instance.
        """
        return handler(self._instance, request, failure)

    def resource(self) -> KleinResource:
        """
        Return an L{IResource} which suitably wraps this app.

        @returns: An L{IResource}
        """

        return KleinResource(self)

    def __get__(self, instance: Any, owner: object) -> Klein:
        """
        Get an instance of L{Klein} bound to C{instance}.
        """
        if instance is None:
            return self

        if self._boundAs is None:
            for name in dir(owner):
                # Properties may raise an AttributeError on access even though
                # they're visible on the instance, we can ignore those because
                # Klein instances won't raise AttributeError.
                obj = getattr(owner, name, None)
                if obj is self:
                    self._boundAs = name
                    break
            else:
                self._boundAs = "unknown_" + str(id(self))

        boundName = f"__klein_bound_{self._boundAs}__"
        k = cast(
            Optional["Klein"], getattr(instance, boundName, lambda: None)()
        )

        if k is None:
            k = self.__class__()
            k._url_map = self._url_map
            k._endpoints = self._endpoints
            k._error_handlers = self._error_handlers
            k._instance = instance
            kref = ref(k)
            try:
                setattr(instance, boundName, kref)
            except AttributeError:
                pass

        return k

    @staticmethod
    def _segments_in_url(url: str) -> int:
        segment_count = url.count("/")
        if url.endswith("/"):
            segment_count -= 1
        return segment_count

    @_normalFunction
    @_routeArgsFrom(_removeStringArg(Rule))
    def route(
        self,
        url: str,
        *args: Any,
        branch: bool = False,
        **kwargs: Any,
    ) -> Callable[[KleinRouteHandlerT], KleinRouteHandlerT]:
        """
        Add a new handler for C{url} passing C{args} and C{kwargs} directly to
        C{werkzeug.routing.Rule}.  The handler function will be passed at least
        one argument an L{twisted.web.server.Request} and any keyword arguments
        taken from the C{url} pattern.

        ::
            @app.route("/")
            def index(request):
                return "Hello"

        @param url: A werkzeug URL pattern given to C{werkzeug.routing.Rule}.
        @param branch: A bool indiciated if a branch endpoint should
            be added that allows all child path segments that don't
            match some other route to be consumed.  Default C{False}.

        @returns: decorated handler function.
        """
        segment_count = self._segments_in_url(url) + self._subroute_segments

        @named("router for '" + url + "'")
        def deco(f: KleinRouteHandlerT) -> KleinRouteHandlerT:
            metadata = route_metadata(f)
            kwargs.setdefault(
                "endpoint",
                metadata.__name__,
            )
            if branch:
                branchKwargs = kwargs.copy()
                branchKwargs["endpoint"] = branchKwargs["endpoint"] + "_branch"

                @modified(f"branch route '{url}' executor", f)
                def branch_f(
                    instance: Any,
                    request: IRequest,
                    *a: Any,
                    **kw: Any,
                ) -> KleinRenderable:
                    IKleinRequest(request).branch_segments = kw.pop(
                        "__rest__", ""
                    ).split("/")
                    return _call(instance, f, request, *a, **kw)

                branch_metadata = route_metadata(branch_f)
                branch_metadata.segment_count = segment_count

                self._endpoints[branchKwargs["endpoint"]] = branch_f
                self._url_map.add(
                    Rule(
                        url.rstrip("/") + "/" + "<path:__rest__>",
                        *args,
                        **branchKwargs,
                    )
                )

            @modified(f"route '{url}' executor", f)
            def _f(
                instance: Any,
                request: IRequest,
                *a: Any,
                **kw: Any,
            ) -> KleinRenderable:
                return _call(instance, f, request, *a, **kw)

            exec_metadata = route_metadata(_f)
            exec_metadata.segment_count = segment_count

            self._endpoints[kwargs["endpoint"]] = _f
            self._url_map.add(Rule(url, *args, **kwargs))
            return f

        return deco

    @contextmanager
    def subroute(self, prefix: str) -> Iterator[Klein]:
        """
        Within this block, C{@route} adds rules to a
        C{werkzeug.routing.Submount}.

        This is implemented by tinkering with the instance's C{_url_map}
        variable. A context manager allows us to gracefully use the pattern of
        "change a variable, do some things with the new value, then put it back
        to how it was before.

        Named "subroute" to try and give callers a better idea of its
        relationship to C{@route}.

        Usage:
        ::
            with app.subroute("/prefix") as app:
                @app.route("/foo")
                def foo_handler(request):
                    return 'I respond to /prefix/foo'

        @param prefix: The string that will be prepended to the paths of all
                       routes established during the with-block.
        """

        _map_before_submount = self._url_map

        segments = self._segments_in_url(prefix)

        class SubmountMap:
            def __init__(self) -> None:
                self.rules: list[Rule] = []

            def add(self, rule: Rule) -> None:
                self.rules.append(rule)

        submount_map = SubmountMap()

        try:
            self._url_map = cast(Map, submount_map)
            self._subroute_segments += segments
            yield self
            _map_before_submount.add(Submount(prefix, submount_map.rules))
        finally:
            self._url_map = _map_before_submount
            self._subroute_segments -= segments

    @overload
    def handle_errors(
        self,
        f_or_exception: KleinErrorHandler,
        *additional_exceptions: Type[Exception],
    ) -> Callable[[KleinErrorHandler], Callable]:
        ...  # pragma: no cover

    @overload
    def handle_errors(
        self,
        f_or_exception: Type[Exception],
        *additional_exceptions: Type[Exception],
    ) -> Callable[[KleinErrorHandler], Callable]:
        ...  # pragma: no cover

    def handle_errors(
        self,
        f_or_exception: Union[KleinErrorHandler, Type[Exception]],
        *additional_exceptions: Type[Exception],
    ) -> Callable[[KleinErrorHandler], Callable]:
        """
        Register an error handler. This decorator supports two syntaxes. The
        simpler of these can be used to register a handler for all C{Exception}
        types::

            @app.handle_errors
            def error_handler(request, failure):
                request.setResponseCode(500)
                return 'Uh oh'

        Alternately, a handler can be registered for one or more specific
        C{Exception} types::

            @app.handle_errors(EncodingError, ValidationError):
            def error_handler(request, failure)
                request.setResponseCode(400)
                return failure.getTraceback()

        The handler will be passed a L{twisted.web.server.Request} as well as a
        L{twisted.python.failure.Failure} instance. Error handlers may return a
        deferred, a failure or a response body.

        If more than one error handler is registered, the handlers will be
        executed in the order in which they are defined, until a handler is
        encountered which completes successfully. If no handler completes
        successfully, L{twisted.web.server.Request}'s processingFailed() method
        will be called.

        In addition to handling errors that occur within a L{KleinRouteHandler},
        error handlers also handle any L{werkzeug.exceptions.HTTPException}
        which is raised during request routing.

        In particular, C{werkzeug.exceptions.NotFound} will be raised if no
        matching route is found, so to return a custom 404 users can do the
        following::

            @app.handle_errors(NotFound)
            def error_handler(request, failure):
                request.setResponseCode(404)
                return 'Not found'

        @param f_or_exception: An error handler function, or an C{Exception}
            subclass to scope the decorated handler to.
        @param additional_exceptions: Additional C{Exception} subclasses to
            scope the decorated function to.

        @returns: decorated error handler function.
        """
        # Try to detect calls using the "simple" @app.handle_error syntax by
        # introspecting the first argument - if it isn't a type which
        # subclasses Exception we assume the simple syntax was used.
        if not isinstance(f_or_exception, type) or not issubclass(
            f_or_exception, Exception
        ):
            # f_or_exception is a KleinErrorHandler
            f = cast(KleinErrorHandler, f_or_exception)
            return self.handle_errors(Exception)(f)

        # f_or_exception is an Exception class
        exceptions = [f_or_exception] + list(additional_exceptions)

        def deco(f: KleinErrorHandler) -> Callable:
            @modified("error handling wrapper", f)
            def _f(
                instance: Optional[Klein],
                request: IRequest,
                failure: Failure,
            ) -> KleinRenderable:
                return _call(instance, f, request, failure)

            self._error_handlers.append((exceptions, _f))

            return cast(Callable, _f)

        return deco

    def urlFor(
        self,
        request: IRequest,
        endpoint: str,
        values: Optional[Mapping[str, KleinQueryValue]] = None,
        method: Optional[str] = None,
        force_external: bool = False,
        append_unknown: bool = True,
    ) -> str:
        host = request.getHeader("host")
        if host is None:
            if force_external:
                raise ValueError(
                    "Cannot build external URL if request"
                    " doesn't contain Host header"
                )
            host = ""
        return buildURL(
            self.url_map.bind(host),
            endpoint,
            values,
            method,
            force_external,
            append_unknown,
        )

    url_for = urlFor

    def run(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        logFile: Optional[IO] = None,
        endpoint_description: Optional[str] = None,
        displayTracebacks: bool = True,
    ) -> None:
        """
        Run a minimal twisted.web server on the specified C{port}, bound to the
        interface specified by C{host} and logging to C{logFile}.

        This function will run the default reactor for your platform and so
        will block the main thread of your application.  It should be the last
        thing your klein application does.

        @param host: The hostname or IP address to bind the listening socket
            to.  "0.0.0.0" will allow you to listen on all interfaces, and
            "127.0.0.1" will allow you to listen on just the loopback
            interface.

        @param port: The TCP port to accept HTTP requests on.

        @param logFile: The file object to log to, by default C{sys.stdout}

        @param endpoint_description: specification of endpoint. Must contain
             protocol, port and interface. May contain other optional arguments,
             e.g. to use SSL: "ssl:443:privateKey=key.pem:certKey=crt.pem"

        @param displayTracebacks: Weather a processing error will result in
            a page displaying the traceback with debugging information or not.
        """
        if logFile is None:
            logFile = sys.stdout

        log.startLogging(logFile)

        if not endpoint_description:
            endpoint_description = f"tcp:port={port}:interface={host}"

        endpoint = serverFromString(reactor, endpoint_description)

        site = Site(self.resource())
        site.displayTracebacks = displayTracebacks

        endpoint.listen(site)
        reactor.run()  # type: ignore[attr-defined]


_globalKleinApp = Klein()

route = _globalKleinApp.route
run = _globalKleinApp.run
subroute = _globalKleinApp.subroute
resource = _globalKleinApp.resource
handle_errors = _globalKleinApp.handle_errors
urlFor = url_for = _globalKleinApp.urlFor



# -*- test-case-name: klein.test.test_session -*-

from typing import Any, Callable, Dict, Optional, Sequence, Type, Union, cast

import attr
from zope.interface import Interface, implementer

from twisted.internet.defer import inlineCallbacks
from twisted.python.components import Componentized
from twisted.python.reflect import qual
from twisted.web.http import UNAUTHORIZED
from twisted.web.iweb import IRequest
from twisted.web.resource import Resource

from .interfaces import (
    EarlyExit,
    IDependencyInjector,
    IRequestLifecycle,
    IRequiredParameter,
    ISession,
    ISessionProcurer,
    ISessionStore,
    NoSuchSession,
    SessionMechanism,
    TooLateForCookies,
)


@implementer(ISessionProcurer)
@attr.s(auto_attribs=True)
class SessionProcurer:
    """
    A L{SessionProcurer} procures a session from a request and a store.

    @ivar _store: The session store to procure a session from.
    @ivar _maxAge: The maximum age (in seconds) of the session cookie.
    @ivar _secureCookie: The name of the cookie to use for sessions protected
        with TLS (i.e. HTTPS).
    @ivar _insecureCookie: The name of the cookie to use for sessions I{not}
        protected with TLS (i.e. HTTP).
    @ivar _cookieDomain: If set, the domain name to restrict the session cookie
        to.
    @ivar _cookiePath: If set, the URL path to restrict the session cookie to.
    @ivar _secureTokenHeader: The name of the HTTPS header to try to extract a
        session token from; API clients should use this header, rather than a
        cookie.
    @ivar _insecureTokenHeader: The name of the HTTP header to try to extract a
        session token from; API clients should use this header, rather than a
        cookie.
    @ivar _setCookieOnGET: Automatically request that the session store create
        a session if one is not already associated with the request and the
        request is a GET.
    """

    _store: ISessionStore

    _maxAge: int = 3600
    _secureCookie: bytes = b"Klein-Secure-Session"
    _insecureCookie: bytes = b"Klein-INSECURE-Session"
    _cookieDomain: Optional[bytes] = None
    _cookiePath: bytes = b"/"

    _secureTokenHeader: bytes = b"X-Auth-Token"
    _insecureTokenHeader: bytes = b"X-INSECURE-Auth-Token"
    _setCookieOnGET: bool = True

    @inlineCallbacks
    def procureSession(
        self, request: IRequest, forceInsecure: bool = False
    ) -> Any:
        alreadyProcured = cast(Componentized, request).getComponent(ISession)
        if alreadyProcured is not None:
            if not forceInsecure or not request.isSecure():
                return alreadyProcured

        if request.isSecure():
            if forceInsecure:
                tokenHeader = self._insecureTokenHeader
                cookieName: Union[str, bytes] = self._insecureCookie
                sentSecurely = False
            else:
                tokenHeader = self._secureTokenHeader
                cookieName = self._secureCookie
                sentSecurely = True
        else:
            # Have we inadvertently disclosed a secure token over an insecure
            # transport, for example, due to a buggy client?
            allPossibleSentTokens: Sequence[str] = sum(
                (
                    request.requestHeaders.getRawHeaders(header, [])
                    for header in [
                        self._secureTokenHeader,
                        self._insecureTokenHeader,
                    ]
                ),
                [],
            ) + [
                it
                for it in [
                    request.getCookie(cookie)
                    for cookie in [self._secureCookie, self._insecureCookie]
                ]
                if it
            ]
            # Does it seem like this check is expensive? It sure is! Don't want
            # to do it? Turn on your dang HTTPS!
            self._store.sentInsecurely(allPossibleSentTokens)
            tokenHeader = self._insecureTokenHeader
            cookieName = self._insecureCookie
            sentSecurely = False
            # Fun future feature: honeypot that does this over HTTPS, but sets
            # isSecure() to return false because it serves up a cert for the
            # wrong hostname or an invalid cert, to keep API clients honest
            # about chain validation.
        sentHeader = (request.getHeader(tokenHeader) or b"").decode("utf-8")
        sentCookie = (request.getCookie(cookieName) or b"").decode("utf-8")
        if sentHeader:
            mechanism = SessionMechanism.Header
        else:
            mechanism = SessionMechanism.Cookie
        if not (sentHeader or sentCookie):
            session = None
        else:
            try:
                session = yield self._store.loadSession(
                    sentHeader or sentCookie, sentSecurely, mechanism
                )
            except NoSuchSession:
                if mechanism == SessionMechanism.Header:
                    raise
                session = None
        if mechanism == SessionMechanism.Cookie and (
            session is None or session.identifier != sentCookie
        ):
            if session is None:
                if request.startedWriting:  # type: ignore[attr-defined]
                    # At this point, if the mechanism is Header, we either have
                    # a valid session or we bailed after NoSuchSession above.
                    raise TooLateForCookies(
                        "You tried initializing a cookie-based session too"
                        " late in the request pipeline; the headers"
                        " were already sent."
                    )
                if request.method != b"GET":
                    # Sessions should only ever be auto-created by GET
                    # requests; there's no way that any meaningful data
                    # manipulation could succeed (no CSRF token check could
                    # ever succeed, for example).
                    raise NoSuchSession(
                        "Can't initialize a session on a "
                        "{method} request.".format(
                            method=request.method.decode("ascii")
                        )
                    )
                if not self._setCookieOnGET:
                    # We don't have a session ID at all, and we're not allowed
                    # by policy to set a cookie on the client.
                    raise NoSuchSession(
                        "Cannot auto-initialize a session for this request."
                    )
                session = yield self._store.newSession(sentSecurely, mechanism)
            identifierInCookie = session.identifier
            if not isinstance(identifierInCookie, str):
                identifierInCookie = identifierInCookie.encode("ascii")
            if not isinstance(cookieName, str):
                cookieName = cookieName.decode("ascii")
            request.addCookie(  # type: ignore[call-arg]
                cookieName,
                identifierInCookie,
                max_age=str(self._maxAge),
                domain=self._cookieDomain,
                path=self._cookiePath,
                secure=sentSecurely,
                httpOnly=True,
            )
        if sentSecurely or not request.isSecure():
            # Do not cache the insecure session on the secure request, thanks.
            cast(Componentized, request).setComponent(ISession, session)
        return session


class AuthorizationDenied(Resource):
    def __init__(self, interface: Type[Interface], instance: Any) -> None:
        self._interface = interface
        super().__init__()

    def render(self, request: IRequest) -> bytes:
        request.setResponseCode(UNAUTHORIZED)
        return f"{qual(self._interface)} DENIED".encode()


@implementer(IDependencyInjector, IRequiredParameter)
@attr.s(auto_attribs=True)
class Authorization:
    """
    Declare that a C{require}-decorated function requires a certain interface
    be authorized from the session.

    This is a dependency injector used in conjunction with a L{klein.Requirer},
    like so::

        from klein import Requirer, SesssionProcurer
        from klein.interfaces import ISession

        from myapp import ISuperDuperAdmin

        requirer = Requirer()
        procurer = SessionProcurer(store=someSessionStore)
        @requirer.prerequisite(ISession)
        def sessionize(request):
            return procurer.procureSession(request)

        app = Klein()

        @requirer.require(
            app.route("/admin"),
            adminPowers=Authorization(ISuperDuperAdmin)
        )
        def myRoute(adminPowers):
            return 'ok admin: ' + adminPowers.doAdminThing()

    In this example, ISuperDuperAdmin is an interface known to your
    application, and (via authorization plugins depending on your session
    storage backend) to your session store.  It has a doAdminThing method.
    When a user hits /admin in their browser, if they are duly authorized,
    they'll see 'ok admin: ' and whatever the super-secret result of
    doAdminThing is.  If not, by default, they'll simply get an HTTP
    UNAUTHORIZED response that says "myapp.ISuperDuperAdmin DENIED".  (This
    behavior can be customized via the C{whenDenied} parameter to
    L{Authorization}.)

    @ivar _interface: the interface that is required.  a provider of this
        interface is what will be dependency-injected.

    @ivar _required: is this authorization required?  If so (the default),
        don't invoke the application code if it cannot be authorized by the
        procured session, and instead return the object specified by whenDenied
        from the dependency-injection process.  If not, then just pass None if
        it is not on the session.

    @ivar _whenDenied: when this authorization is denied, what object - usually
        an IResource - should be returned to the route decorator that was
        passed to L{Requirer.require}?  Note that this will never be used if
        C{required} is set to C{False}.
    """

    _interface: Type[Interface]
    _required: bool = True
    _whenDenied: Callable[[Type[Interface], Any], Any] = AuthorizationDenied

    def registerInjector(
        self,
        injectionComponents: Componentized,
        parameterName: str,
        lifecycle: IRequestLifecycle,
    ) -> IDependencyInjector:
        """
        Register this authorization to inject a parameter.
        """
        return self

    @inlineCallbacks
    def injectValue(
        self, instance: Any, request: IRequest, routeParams: Dict[str, Any]
    ) -> Any:
        """
        Inject a value by asking the request's session.
        """
        # TODO: this could be optimized to do fewer calls to 'authorize' by
        # collecting all the interfaces that are necessary and then using
        # addBeforeHook; the interface would not need to change.
        session = ISession(request)
        provider = (yield session.authorize([self._interface])).get(
            self._interface
        )
        if self._required and provider is None:
            raise EarlyExit(self._whenDenied(self._interface, instance))
        # TODO: CSRF protection should probably go here
        return provider

    def finalize(self) -> None:
        """
        Nothing to finalize when registering.
        """



# -*- test-case-name: klein.test.test_message -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
HTTP message API.
"""

from typing import Any, Optional, Union

from attr import attrib, attrs
from attr.validators import instance_of, optional
from tubes.itube import IFount

from ._imessage import FountAlreadyAccessedError
from ._tubes import bytesToFount, fountToBytes


__all__ = ()


InternalBody = Union[bytes, IFount]


@attrs(frozen=False)
class MessageState:
    """
    Internal mutable state for HTTP message implementations in L{klein}.
    """

    cachedBody: Optional[bytes] = attrib(
        validator=optional(instance_of(bytes)), default=None, init=False
    )

    fountExhausted: bool = attrib(
        validator=instance_of(bool), default=False, init=False
    )


def validateBody(instance: Any, attribute: Any, body: InternalBody) -> None:
    """
    Validator for L{InternalBody}.
    """

    if not isinstance(body, bytes) and not IFount.providedBy(body):
        raise TypeError("body must be bytes or IFount")


def bodyAsFount(body: InternalBody, state: MessageState) -> IFount:
    """
    Return a fount for a given L{InternalBody}.
    """

    if state.fountExhausted:
        raise FountAlreadyAccessedError()
    state.fountExhausted = True

    if isinstance(body, bytes):
        return bytesToFount(body)

    # assuming: IFount.providedBy(body)

    return body


async def bodyAsBytes(body: InternalBody, state: MessageState) -> bytes:
    """
    Return bytes for a given L{InternalBody}.
    """

    if isinstance(body, bytes):
        return body

    # assuming: IFount.providedBy(body)

    if state.cachedBody is not None:
        return state.cachedBody

    state.cachedBody = await fountToBytes(body)

    return state.cachedBody



# -*- test-case-name: klein.test.test_response -*-
# Copyright (c) 2011-2021. See LICENSE for details.

"""
Tests for L{klein._response}.
"""

from .._headers import FrozenHTTPHeaders
from .._imessage import IHTTPMessage
from .._response import FrozenHTTPResponse, IHTTPResponse
from ._trial import TestCase
from .test_message import FrozenHTTPMessageTestsMixIn


__all__ = ()


class FrozenHTTPResponseTests(FrozenHTTPMessageTestsMixIn, TestCase):
    """
    Tests for L{FrozenHTTPResponse}.
    """

    @staticmethod
    def responseFromBytes(data: bytes = b"") -> FrozenHTTPResponse:
        return FrozenHTTPResponse(
            status=200,
            headers=FrozenHTTPHeaders(rawHeaders=()),
            body=data,
        )

    @classmethod
    def messageFromBytes(cls, data: bytes = b"") -> IHTTPMessage:
        return cls.responseFromBytes(data)

    def test_interface(self) -> None:
        """
        L{FrozenHTTPResponse} implements L{IHTTPResponse}.
        """
        response = self.responseFromBytes()
        self.assertProvides(IHTTPResponse, response)

    def test_initInvalidBodyType(self) -> None:
        """
        L{FrozenHTTPResponse} raises L{TypeError} when given a body of an
        unknown type.
        """
        e = self.assertRaises(
            TypeError,
            FrozenHTTPResponse,
            status=200,
            headers=FrozenHTTPHeaders(rawHeaders=()),
            body=object(),
        )
        self.assertEqual(str(e), "body must be bytes or IFount")



# Copyright (c) 2011-2021. See LICENSE for details.

"""
Extensions to Tubes.
"""

from io import BytesIO
from typing import Any, BinaryIO

from attr import attrib, attrs
from attr.validators import instance_of, optional, provides
from tubes.itube import IDrain, IFount, ISegment
from tubes.kit import Pauser, beginFlowingTo
from tubes.undefer import fountToDeferred
from zope.interface import implementer

from twisted.python.failure import Failure


__all__ = ()


# See https://github.com/twisted/tubes/issues/60
async def fountToBytes(fount: IFount) -> bytes:
    chunks = await fountToDeferred(fount)
    return b"".join(chunks)


# See https://github.com/twisted/tubes/issues/60
def bytesToFount(data: bytes) -> IFount:
    return IOFount(source=BytesIO(data))


# https://github.com/twisted/tubes/issues/61
@implementer(IFount)
@attrs(frozen=False)
class IOFount:
    """
    Fount that reads from a file-like-object.
    """

    outputType = ISegment

    _source: BinaryIO = attrib()

    drain: IDrain = attrib(
        validator=optional(provides(IDrain)), default=None, init=False
    )
    _paused = attrib(validator=instance_of(bool), default=False, init=False)

    def __attrs_post_init__(self) -> None:
        self._pauser = Pauser(self._pause, self._resume)

    def _flowToDrain(self) -> None:
        if self.drain is not None and not self._paused:
            data = self._source.read()
            if data:
                self.drain.receive(data)
            self.drain.flowStopped(Failure(StopIteration()))

    def flowTo(self, drain: IDrain) -> IFount:
        result = beginFlowingTo(self, drain)
        self._flowToDrain()
        return result

    def pauseFlow(self) -> Any:
        return self._pauser.pause()

    def stopFlow(self) -> Any:
        return self._pauser.resume()

    def _pause(self) -> None:
        self._paused = True

    def _resume(self) -> None:
        self._paused = False
        self._flowToDrain()



# -*- test-case-name: klein.test.test_memory -*-
from binascii import hexlify
from os import urandom
from typing import Any, Callable, Dict, Iterable, Type, cast

import attr
from zope.interface import Interface, implementer

from twisted.internet.defer import Deferred, fail, succeed
from twisted.python.components import Componentized

from klein.interfaces import (
    ISession,
    ISessionStore,
    NoSuchSession,
    SessionMechanism,
)


_authCB = Callable[[Type[Interface], ISession, Componentized], Any]


@implementer(ISession)
@attr.s(auto_attribs=True)
class MemorySession:
    """
    An in-memory session.
    """

    identifier: str
    isConfidential: bool
    authenticatedBy: SessionMechanism
    _authorizationCallback: _authCB
    _components: Componentized = attr.ib(factory=Componentized)

    def authorize(self, interfaces: Iterable[Type[Interface]]) -> Deferred:
        """
        Authorize each interface by calling back to the session store's
        authorization callback.
        """
        result = {}
        for interface in interfaces:
            provider = self._authorizationCallback(
                interface, self, self._components
            )
            if provider is not None:
                result[interface] = provider
        return succeed(result)


class _MemoryAuthorizerFunction:
    """
    Type shadow for function with the given attribute.
    """

    __memoryAuthInterface__: Type[Interface] = None  # type: ignore[assignment]

    def __call__(
        self, interface: Type[Interface], session: ISession, data: Componentized
    ) -> Any:
        """
        Return a provider of the given interface.
        """


_authFn = Callable[[Type[Interface], ISession, Componentized], Any]


def declareMemoryAuthorizer(
    forInterface: Type[Interface],
) -> Callable[[Callable], _MemoryAuthorizerFunction]:
    """
    Declare that the decorated function is an authorizer usable with a memory
    session store.
    """

    def decorate(decoratee: _authFn) -> _MemoryAuthorizerFunction:
        decoratee = cast(_MemoryAuthorizerFunction, decoratee)
        decoratee.__memoryAuthInterface__ = forInterface
        return decoratee

    return decorate


def _noAuthorization(
    interface: Type[Interface], session: ISession, data: Componentized
) -> None:
    return None


@implementer(ISessionStore)
@attr.s(auto_attribs=True)
class MemorySessionStore:
    authorizationCallback: _authFn = _noAuthorization
    _secureStorage: Dict[str, Any] = attr.ib(factory=dict)
    _insecureStorage: Dict[str, Any] = attr.ib(factory=dict)

    @classmethod
    def fromAuthorizers(
        cls, authorizers: Iterable[_MemoryAuthorizerFunction]
    ) -> "MemorySessionStore":
        """
        Create a L{MemorySessionStore} from a collection of callbacks which can
        do authorization.
        """
        interfaceToCallable = {}
        for authorizer in authorizers:
            assert authorizer.__memoryAuthInterface__ is not None
            specifiedInterface = authorizer.__memoryAuthInterface__
            interfaceToCallable[specifiedInterface] = authorizer

        def authorizationCallback(
            interface: Type[Interface], session: ISession, data: Componentized
        ) -> Any:
            return interfaceToCallable.get(interface, _noAuthorization)(
                interface, session, data
            )

        return cls(authorizationCallback)

    def _storage(self, isConfidential: bool) -> Dict[str, Any]:
        """
        Return the storage appropriate to the isConfidential flag.
        """
        if isConfidential:
            return self._secureStorage
        else:
            return self._insecureStorage

    def newSession(
        self, isConfidential: bool, authenticatedBy: SessionMechanism
    ) -> Deferred:
        storage = self._storage(isConfidential)
        identifier = hexlify(urandom(32)).decode("ascii")
        session = MemorySession(
            identifier,
            isConfidential,
            authenticatedBy,
            self.authorizationCallback,
        )
        storage[identifier] = session
        return succeed(session)

    def loadSession(
        self,
        identifier: str,
        isConfidential: bool,
        authenticatedBy: SessionMechanism,
    ) -> Deferred:
        storage = self._storage(isConfidential)
        if identifier in storage:
            return succeed(storage[identifier])
        else:
            return fail(
                NoSuchSession(
                    "Session not found in memory store {id!r}".format(
                        id=identifier
                    )
                )
            )

    def sentInsecurely(self, tokens: Iterable[str]) -> None:
        return



"""
Tests for L{klein._session}.
"""

from typing import Any, Generator, List, Tuple, Type

from treq.testing import StubTreq
from zope.interface import Interface, implementer

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python.components import Componentized
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.iweb import IRequest

from klein import Authorization, Klein, Requirer, SessionProcurer
from klein.interfaces import ISession, NoSuchSession, TooLateForCookies
from klein.storage.memory import MemorySessionStore, declareMemoryAuthorizer


Sessions = List[ISession]
Errors = List[NoSuchSession]


class ISimpleTest(Interface):
    """
    Interface for testing.
    """

    def doTest() -> int:
        """
        Test method.
        """


class IDenyMe(Interface):
    """
    Interface that is never provided.
    """


@implementer(ISimpleTest)
class SimpleTest:
    """
    Implementation of L{ISimpleTest} for testing.
    """

    def doTest(self) -> int:
        """
        Implementation of L{ISimpleTest}.  Returns 3.
        """
        return 3


@declareMemoryAuthorizer(ISimpleTest)
def memoryAuthorizer(
    interface: Type[Interface], session: ISession, data: Componentized
) -> SimpleTest:
    """
    Authorize the ISimpleTest interface; it always works.
    """
    return SimpleTest()


def simpleSessionRouter() -> Tuple[Sessions, Errors, str, str, StubTreq]:
    """
    Construct a simple router.
    """
    sessions: Sessions = []
    exceptions: Errors = []
    mss = MemorySessionStore.fromAuthorizers([memoryAuthorizer])
    router = Klein()
    token = "X-Test-Session-Token"
    cookie = "X-Test-Session-Cookie"
    sproc = SessionProcurer(
        mss,
        secureTokenHeader=b"X-Test-Session-Token",
        secureCookie=b"X-Test-Session-Cookie",
    )

    @router.route("/")
    @inlineCallbacks
    def route(request: IRequest) -> Generator[Any, object, bytes]:
        try:
            sessions.append(
                (yield sproc.procureSession(request)),  # type: ignore[arg-type]
            )
        except NoSuchSession as nss:
            exceptions.append(nss)
        return b"ok"

    requirer = Requirer()

    @requirer.prerequisite([ISession])
    def procure(request: IRequest) -> Deferred:
        return sproc.procureSession(request)

    @requirer.require(router.route("/test"), simple=Authorization(ISimpleTest))
    def testRoute(simple: SimpleTest) -> str:
        return "ok: " + str(simple.doTest() + 4)

    @requirer.require(router.route("/denied"), nope=Authorization(IDenyMe))
    def testDenied(nope: IDenyMe) -> str:
        return "bad"

    treq = StubTreq(router.resource())
    return sessions, exceptions, token, cookie, treq


class ProcurementTests(SynchronousTestCase):
    """
    Tests for L{klein.SessionProcurer}.
    """

    def test_procurementSecurity(self) -> None:
        """
        Once a session is negotiated, it should be the identical object to
        avoid duplicate work - unless we are using forceInsecure to retrieve
        the insecure session from a secure request, in which case the result
        should not be cached.
        """
        sessions = []
        mss = MemorySessionStore()
        router = Klein()

        @router.route("/")
        @inlineCallbacks
        def route(request: IRequest) -> Generator[Any, object, bytes]:
            sproc = SessionProcurer(mss)
            sessions.append((yield sproc.procureSession(request)))
            sessions.append((yield sproc.procureSession(request)))
            sessions.append(
                (yield sproc.procureSession(request, forceInsecure=True))
            )
            return b"sessioned"

        treq = StubTreq(router.resource())
        self.successResultOf(treq.get("http://unittest.example.com/"))
        self.assertIs(sessions[0], sessions[1])
        self.assertIs(sessions[0], sessions[2])
        self.successResultOf(treq.get("https://unittest.example.com/"))
        self.assertIs(sessions[3], sessions[4])
        self.assertIsNot(sessions[3], sessions[5])

    def test_procuredTooLate(self) -> None:
        """
        If you start writing stuff to the response before procuring the
        session, when cookies need to be set, you will get a comprehensible
        error.
        """
        mss = MemorySessionStore()
        router = Klein()

        @router.route("/")
        @inlineCallbacks
        def route(request: IRequest) -> Generator[Any, object, None]:
            sproc = SessionProcurer(mss)
            request.write(b"oops...")
            with self.assertRaises(TooLateForCookies):
                yield sproc.procureSession(request)
            request.write(b"bye")
            request.finish()

        treq = StubTreq(router.resource())
        result = self.successResultOf(treq.get("http://unittest.example.com/"))
        self.assertEqual(self.successResultOf(result.content()), b"oops...bye")

    def test_cookiesTurnedOff(self) -> None:
        """
        If cookies can't be set, then C{procureSession} raises
        L{NoSuchSession}.
        """
        mss = MemorySessionStore()
        router = Klein()

        @router.route("/")
        @inlineCallbacks
        def route(request: IRequest) -> Generator[Any, object, bytes]:
            sproc = SessionProcurer(mss, setCookieOnGET=False)
            with self.assertRaises(NoSuchSession):
                yield sproc.procureSession(request)
            return b"no session"

        treq = StubTreq(router.resource())
        result = self.successResultOf(treq.get("http://unittest.example.com/"))
        self.assertEqual(self.successResultOf(result.content()), b"no session")

    def test_unknownSessionHeader(self) -> None:
        """
        Unknown session IDs in auth headers will be immediately rejected with
        L{NoSuchSession}.
        """
        sessions, exceptions, token, cookie, treq = simpleSessionRouter()

        response = self.successResultOf(
            treq.get("https://unittest.example.com/", headers={token: "bad"})
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(len(sessions), 0)
        self.assertEqual(len(exceptions), 1)

    def test_unknownSessionCookieGET(self) -> None:
        """
        Unknown session IDs in cookies will result in a new session being
        created.
        """
        badSessionID = "bad"
        sessions, exceptions, token, cookie, treq = simpleSessionRouter()
        response = self.successResultOf(
            treq.get(
                "https://unittest.example.com/", cookies={cookie: badSessionID}
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(len(exceptions), 0)
        self.assertEqual(len(sessions), 1)
        self.assertNotEqual(sessions[0].identifier, badSessionID)

    def test_unknownSessionCookiePOST(self) -> None:
        """
        Unknown session IDs in cookies for POST requests will result in a
        NoSuchSession error.
        """
        badSessionID = "bad"
        sessions, exceptions, token, cookie, treq = simpleSessionRouter()
        response = self.successResultOf(
            treq.post(
                "https://unittest.example.com/", cookies={cookie: badSessionID}
            )
        )
        self.assertEqual(response.code, 200)
        self.assertEqual(len(exceptions), 1)
        self.assertEqual(len(sessions), 0)

    def test_authorization(self) -> None:
        """
        When L{Requirer.require} is used with L{Authorization} and the session
        knows how to supply that authorization, it is passed to the object.
        """
        sessions, exceptions, token, cookie, treq = simpleSessionRouter()
        response = self.successResultOf(
            treq.get("https://unittest.example.com/test")
        )
        self.assertEqual(self.successResultOf(response.content()), b"ok: 7")

    def test_authorizationDenied(self) -> None:
        """
        When L{Requirer.require} is used with an L{Authorization} and the
        session does I{not} know how to supply that authorization, the callable
        is not invoked.
        """
        sessions, exceptions, token, cookie, treq = simpleSessionRouter()
        response = self.successResultOf(
            treq.get("https://unittest.example.com/denied")
        )
        self.assertEqual(
            self.successResultOf(response.content()),
            b"klein.test.test_session.IDenyMe DENIED",
        )



# -*- test-case-name: klein.test.test_resource -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Sequence, Tuple, Union, cast

from werkzeug.exceptions import HTTPException

from twisted.internet import defer
from twisted.internet.defer import Deferred, maybeDeferred, succeed
from twisted.python import log
from twisted.python.failure import Failure
from twisted.web import server
from twisted.web.iweb import IRenderable, IRequest
from twisted.web.resource import IResource, Resource, getChildForRequest
from twisted.web.template import renderElement

from ._dihttp import Response
from ._interfaces import IKleinRequest


if TYPE_CHECKING:
    # NB: circular import, must not be imported at runtime.
    from ._app import (
        ErrorHandlers,
        Klein,
        KleinRenderable,
        KleinRouteHandler,
        RouteMetadata,
    )


def route_metadata(handler: KleinRouteHandler) -> RouteMetadata:
    return handler  # type:ignore[return-value]


def ensure_utf8_bytes(v: Union[str, bytes]) -> bytes:
    """
    Coerces a value which is either a C{str} or C{bytes} to a C{bytes}.
    If ``v`` is a C{str} object it is encoded as utf-8.
    """
    if isinstance(v, str):
        v = v.encode("utf-8")
    return v


class _StandInResource:
    """
    A standin for a Resource.

    This is a sentinel value for L{KleinResource}, to say that we are rendering
    a L{Resource}, which may close the connection itself later.
    """


StandInResource = cast("KleinResource", _StandInResource())


class URLDecodeError(Exception):
    """
    Raised if one or more string parts of the URL could not be decoded.
    """

    __slots__ = ["errors"]

    def __init__(self, errors: Sequence[Tuple[str, Failure]]) -> None:
        """
        @param errors: Sequence of decoding errors, expressed as tuples
            of names and an associated failure.
        """
        self.errors = errors

    def __repr__(self) -> str:
        return f"<URLDecodeError(errors={self.errors!r})>"


def extractURLparts(request: IRequest) -> Tuple[str, str, int, str, str]:
    """
    Extracts and decodes URI parts from C{request}.

    All strings must be UTF8-decodable.

    @param request: A Twisted Web request.

    @raise URLDecodeError: If one of the parts could not be decoded as UTF-8.

    @return: L{tuple} of the URL scheme, the server name, the server port, the
        path info and the script name.
    """
    server_name = request.getRequestHostname()
    if hasattr(request.getHost(), "port"):
        server_port = request.getHost().port
    else:
        server_port = 0
    if (bool(request.isSecure()), server_port) not in [
        (True, 443),
        (False, 80),
        (False, 0),
        (True, 0),
    ]:
        server_name = b"%s:%d" % (server_name, server_port)

    script_name = b""
    if request.prepath:
        script_name = b"/".join(request.prepath)

        if not script_name.startswith(b"/"):
            script_name = b"/" + script_name

    path_info = b""
    if request.postpath:
        path_info = b"/".join(request.postpath)

        if not path_info.startswith(b"/"):
            path_info = b"/" + path_info

    url_scheme = "https" if request.isSecure() else "http"

    utf8Failures = []
    try:
        server_name = server_name.decode("utf-8")
    except UnicodeDecodeError:
        utf8Failures.append(("SERVER_NAME", Failure()))
    try:
        path_text = path_info.decode("utf-8")
    except UnicodeDecodeError:
        utf8Failures.append(("PATH_INFO", Failure()))
    try:
        script_text = script_name.decode("utf-8")
    except UnicodeDecodeError:
        utf8Failures.append(("SCRIPT_NAME", Failure()))

    if utf8Failures:
        raise URLDecodeError(utf8Failures)

    return url_scheme, server_name, server_port, path_text, script_text


class KleinResource(Resource):
    """
    A ``Resource`` that can do URL routing.
    """

    isLeaf = True

    def __init__(self, app: Klein) -> None:
        super().__init__()
        self._app = app

    def __eq__(self, other: object) -> bool:
        if isinstance(other, KleinResource):
            return vars(self) == vars(other)
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    def render(self, request: IRequest) -> KleinRenderable:
        # Stuff we need to know for the mapper.
        try:
            (
                url_scheme,
                server_name,
                server_port,
                path_info,
                script_name,
            ) = extractURLparts(request)
        except URLDecodeError as e:
            for what, fail in e.errors:
                log.err(fail, f"Invalid encoding in {what}.")
            request.setResponseCode(400)
            return b"Non-UTF-8 encoding in URL."

        # Bind our mapper.
        mapper = self._app.url_map.bind(
            server_name,
            script_name,
            path_info=path_info,
            default_method=request.method.decode("utf-8"),
            url_scheme=url_scheme,
        )
        # Make the mapper available to the view.
        kleinRequest = IKleinRequest(request)
        kleinRequest.mapper = mapper

        # Make sure we'll notice when the connection goes away unambiguously.
        request_finished = [False]

        def _finish(result: object) -> None:
            request_finished[0] = True

        def _execute() -> Deferred:
            # Actually doing the match right here. This can cause an exception
            # to percolate up. If that happens it will be handled below in
            # processing_failed, either by a user-registered error handler or
            # one of our defaults.
            (rule, kwargs) = mapper.match(return_rule=True)
            endpoint = rule.endpoint

            # Try pretty hard to fix up prepath and postpath.
            segment_count = route_metadata(
                self._app.endpoints[endpoint]
            ).segment_count
            request.prepath.extend(request.postpath[:segment_count])
            request.postpath = request.postpath[segment_count:]

            request.notifyFinish().addBoth(  # type: ignore[attr-defined]
                _finish,
            )

            # Standard Twisted Web stuff. Defer the method action, giving us
            # something renderable or printable. Return NOT_DONE_YET and set up
            # the incremental renderer.
            d = maybeDeferred(
                self._app.execute_endpoint, endpoint, request, **kwargs
            )

            request.notifyFinish().addErrback(  # type: ignore[attr-defined]
                lambda _: d.cancel(),
            )

            return d

        d = maybeDeferred(_execute)

        # typing note: returns Any because Response._applyToRequest returns Any
        def process(r: object) -> Any:
            """
            Recursively go through r and any child Resources until something
            returns an IRenderable, then render it and let the result of that
            bubble back up.
            """
            if isinstance(r, Response):
                r = r._applyToRequest(request)

            if IResource.providedBy(r):
                request.render(  # type: ignore[attr-defined]
                    getChildForRequest(r, request)
                )
                return StandInResource

            if IRenderable.providedBy(r):
                renderElement(request, r)
                return StandInResource

            return r

        d.addCallback(process)

        def processing_failed(
            failure: Failure, error_handlers: ErrorHandlers
        ) -> Optional[Deferred]:
            # The failure processor writes to the request.  If the
            # request is already finished we should suppress failure
            # processing.  We don't return failure here because there
            # is no way to surface this failure to the user if the
            # request is finished.
            if request_finished[0]:
                if not failure.check(defer.CancelledError):
                    log.err(failure, "Unhandled Error Processing Request.")
                return None

            # If there are no more registered handlers, apply some defaults
            if len(error_handlers) == 0:
                if failure.check(HTTPException):
                    he = failure.value
                    assert isinstance(he, HTTPException)
                    request.setResponseCode(he.code)
                    resp = he.get_response({})

                    for header, value in resp.headers:
                        request.setHeader(
                            ensure_utf8_bytes(header), ensure_utf8_bytes(value)
                        )

                    encoded = resp.iter_encoded()  # type: ignore[attr-defined]
                    return succeed(ensure_utf8_bytes(b"".join(encoded)))
                else:
                    request.processingFailed(  # type: ignore[attr-defined]
                        failure,
                    )
                    return None

            error_handler = error_handlers[0]

            # Each error handler is a tuple of
            # (list_of_exception_types, handler_fn)
            if failure.check(*error_handler[0]):
                d = maybeDeferred(
                    self._app.execute_error_handler,
                    error_handler[1],  # type: ignore[arg-type]
                    request,
                    failure,
                )

                d.addCallback(process)

                return d.addErrback(processing_failed, error_handlers[1:])

            return processing_failed(failure, error_handlers[1:])

        d.addErrback(processing_failed, self._app._error_handlers)

        def write_response(
            r: Union[_StandInResource, str, bytes, int, None]
        ) -> None:
            if r is StandInResource:
                return

            if isinstance(r, str):
                r = r.encode("utf-8")

            if isinstance(r, bytes):
                request.write(r)

            if not request_finished[0]:
                request.finish()

        d.addCallback(write_response)
        d.addErrback(log.err, _why="Unhandled Error writing response")

        return server.NOT_DONE_YET  # type: ignore[return-value]
